from pynwb.testing.mock.file import mock_Subject, mock_NWBFile

from pynwb.testing.mock.base import mock_TimeSeries

from pynwb.testing.mock.ophys import (
    mock_ImagingPlane,
    mock_OnePhotonSeries,
    mock_TwoPhotonSeries,
    mock_RoiResponseSeries,
    mock_PlaneSegmentation,
    mock_OpticalChannel,
    mock_Fluorescence,
    mock_DfOverF,
    mock_ImageSegmentation,
)

from pynwb.testing.mock.ogen import (
    mock_OptogeneticStimulusSite,
    mock_OptogeneticSeries
)

from pynwb.testing.mock.device import mock_Device

from pynwb.testing.mock.behavior import (
    mock_Position,
    mock_PupilTracking,
    mock_CompassDirection,
    mock_SpatialSeries,
)

from pynwb.testing.mock.ecephys import (
    mock_ElectrodeGroup,
    mock_ElectrodeTable,
    mock_ElectricalSeries,
    mock_SpikeEventSeries,
)

from pynwb.testing.mock.icephys import (
    mock_IntracellularElectrode,
    mock_CurrentClampSeries,
    mock_IZeroClampSeries,
    mock_VoltageClampSeries,
    mock_VoltageClampStimulusSeries,
    mock_CurrentClampStimulusSeries,
    mock_IntracellularRecordingsTable,
)

import pytest

from pynwb.testing.mock.utils import name_generator, name_generator_registry


@pytest.mark.parametrize(
    "mock_function", [
        mock_ImagingPlane,
        mock_OnePhotonSeries,
        mock_TwoPhotonSeries,
        mock_RoiResponseSeries,
        mock_PlaneSegmentation,
        mock_OpticalChannel,
        mock_Fluorescence,
        mock_DfOverF,
        mock_ImageSegmentation,
        mock_OptogeneticStimulusSite,
        mock_OptogeneticSeries,
        mock_Device,
        mock_Position,
        mock_PupilTracking,
        mock_CompassDirection,
        mock_SpatialSeries,
        mock_ElectrodeGroup,
        mock_ElectrodeTable,
        mock_ElectricalSeries,
        mock_SpikeEventSeries,
        mock_Subject,
        mock_NWBFile,
        mock_TimeSeries,
        mock_CurrentClampSeries,
        mock_IZeroClampSeries,
        mock_VoltageClampSeries,
        mock_VoltageClampStimulusSeries,
        mock_IntracellularElectrode,
        mock_CurrentClampStimulusSeries,
        mock_IntracellularRecordingsTable,
    ],
)
def test_mock(mock_function):
    mock_function()


def test_name_generator():

    name_generator_registry.clear()  # reset registry

    assert name_generator("TimeSeries") == "TimeSeries"
    assert name_generator("TimeSeries") == "TimeSeries2"



from datetime import datetime
import numpy as np
from pathlib import Path
from pynwb import NWBFile, NWBHDF5IO, __version__, TimeSeries, get_class, load_namespaces
from pynwb.file import Subject
from pynwb.image import ImageSeries
from pynwb.spec import NWBNamespaceBuilder, export_spec, NWBGroupSpec, NWBAttributeSpec


# pynwb 1.0.2 should be installed with hdmf 1.0.3
# pynwb 1.0.3 should be installed with hdmf 1.0.5
# pynwb 1.1.0 should be installed with hdmf 1.2.0
# pynwb 1.1.1+ should be installed with an appropriate version of hdmf


def _write(test_name, nwbfile):
    filename = str("%s/%s_%s.nwb") % (Path(__file__).parent, __version__, test_name)
    with NWBHDF5IO(filename, 'w') as io:
        io.write(nwbfile)

    return filename


def _make_empty():
    nwbfile = NWBFile(session_description='ADDME',
                      identifier='ADDME',
                      session_start_time=datetime.now().astimezone())
    test_name = 'nwbfile'
    _write(test_name, nwbfile)


def _make_str_experimenter():
    nwbfile = NWBFile(session_description='ADDME',
                      identifier='ADDME',
                      session_start_time=datetime.now().astimezone(),
                      experimenter='one experimenter')
    test_name = 'str_experimenter'
    _write(test_name, nwbfile)


def _make_str_pub():
    nwbfile = NWBFile(session_description='ADDME',
                      identifier='ADDME',
                      session_start_time=datetime.now().astimezone(),
                      related_publications='one publication')
    test_name = 'str_pub'
    _write(test_name, nwbfile)


def _make_timeseries_no_data():
    nwbfile = NWBFile(session_description='ADDME',
                      identifier='ADDME',
                      session_start_time=datetime.now().astimezone())
    ts = TimeSeries(
        name='test_timeseries',
        rate=1.,
        unit='unit',
    )
    nwbfile.add_acquisition(ts)

    test_name = 'timeseries_no_data'
    _write(test_name, nwbfile)


def _make_timeseries_no_unit():
    nwbfile = NWBFile(session_description='ADDME',
                      identifier='ADDME',
                      session_start_time=datetime.now().astimezone())
    ts = TimeSeries(
        name='test_timeseries',
        data=[0],
        rate=1.,
    )
    nwbfile.add_acquisition(ts)

    test_name = 'timeseries_no_unit'
    _write(test_name, nwbfile)


def _make_imageseries_no_data():
    """Create a test file with an ImageSeries with no data."""
    nwbfile = NWBFile(session_description='ADDME',
                      identifier='ADDME',
                      session_start_time=datetime.now().astimezone())
    image_series = ImageSeries(
        name='test_imageseries',
        external_file=['external_file'],
        starting_frame=[1],
        format='external',
        timestamps=[1., 2., 3.]
    )

    nwbfile.add_acquisition(image_series)

    test_name = 'imageseries_no_data'
    _write(test_name, nwbfile)


def _make_imageseries_no_unit():
    """Create a test file with an ImageSeries with data and no unit."""
    nwbfile = NWBFile(session_description='ADDME',
                      identifier='ADDME',
                      session_start_time=datetime.now().astimezone())
    image_series = ImageSeries(
        name='test_imageseries',
        data=np.ones((3, 3, 3)),
        timestamps=[1., 2., 3.]
    )

    nwbfile.add_acquisition(image_series)

    test_name = 'imageseries_no_unit'
    _write(test_name, nwbfile)


def _make_imageseries_non_external_format():
    """Create a test file with an ImageSeries with external_file set and format != 'external'."""
    nwbfile = NWBFile(session_description='ADDME',
                      identifier='ADDME',
                      session_start_time=datetime.now().astimezone())
    image_series = ImageSeries(
        name='test_imageseries',
        external_file=['external_file'],
        starting_frame=[1],
        format='tiff',
        timestamps=[1., 2., 3.]
    )

    nwbfile.add_acquisition(image_series)

    test_name = 'imageseries_non_external_format'
    _write(test_name, nwbfile)


def _make_imageseries_nonmatch_starting_frame():
    """Create a test file with an ImageSeries where len(starting_frame) != len(external_file)."""
    nwbfile = NWBFile(session_description='ADDME',
                      identifier='ADDME',
                      session_start_time=datetime.now().astimezone())
    image_series = ImageSeries(
        name='test_imageseries',
        external_file=['external_file'],
        starting_frame=[1, 2, 3],
        format='external',
        timestamps=[1., 2., 3.]
    )

    nwbfile.add_acquisition(image_series)

    test_name = 'imageseries_nonmatch_starting_frame'
    _write(test_name, nwbfile)


def _make_empty_with_extension():
    ns_builder = NWBNamespaceBuilder(
        doc="An NWB test extension",
        name="ndx-testextension",
        version="0.1.0",
        author="PyNWB Test File Generator",
        contact="my_email@example.com",
    )

    ns_builder.include_type('TimeSeries', namespace='core')
    tetrode_series = NWBGroupSpec(
        neurodata_type_def='TimeSeriesWithID',
        neurodata_type_inc='TimeSeries',
        doc=('An extension of TimeSeries to include an ID.'),
        attributes=[
            NWBAttributeSpec(
                name='id',
                doc='The time series ID.',
                dtype='int32'
            ),
        ],
    )

    new_data_types = [tetrode_series]

    # export the spec to yaml files in the current directory
    export_spec(ns_builder, new_data_types, output_dir=".")

    nwbfile = NWBFile(session_description='ADDME',
                      identifier='ADDME',
                      session_start_time=datetime.now().astimezone())

    load_namespaces("ndx-testextension.namespace.yaml")  # load from the current directory
    TimeSeriesWithID = get_class("TimeSeriesWithID", "ndx-testextension")
    ts = TimeSeriesWithID(
        name="test_ts",
        data=[1., 2., 3.],
        description="ADDME",
        unit="ADDME",
        rate=1.,
        id=1,
    )
    nwbfile.add_acquisition(ts)
    test_name = 'nwbfile_with_extension'
    _write(test_name, nwbfile)


def _make_subject_without_age_reference():
    """Create a test file without a value for age_reference."""
    nwbfile = NWBFile(session_description='ADDME',
                      identifier='ADDME',
                      session_start_time=datetime.now().astimezone())
    subject = Subject(
        age="P90D",
        description="A rat",
        subject_id="RAT123",
    )

    nwbfile.subject = subject

    test_name = 'subject_no_age__reference'
    _write(test_name, nwbfile)


if __name__ == '__main__':
    # install these versions of PyNWB and run this script to generate new files
    # python src/pynwb/testing/make_test_files.py
    # files will be made in src/pynwb/testing/
    # files should be moved to tests/back_compat/

    # NOTE: this script is run in the GitHub Actions workflow generate_test_files.yml

    if __version__ == '1.1.2':
        _make_empty()
        _make_str_experimenter()
        _make_str_pub()

    if __version__ == '1.5.1':
        _make_timeseries_no_data()
        _make_timeseries_no_unit()
        _make_imageseries_no_data()
        _make_imageseries_no_unit()

    if __version__ == '2.1.0':
        _make_imageseries_no_data()
        _make_imageseries_non_external_format()
        _make_imageseries_nonmatch_starting_frame()
        _make_empty_with_extension()

    if __version__ == "2.2.0":
        _make_subject_without_age_reference()



import subprocess
import re
from unittest.mock import patch
from io import StringIO

from pynwb.testing import TestCase
from pynwb import validate, NWBHDF5IO


class TestValidateCLI(TestCase):

    # 1.0.2_nwbfile.nwb has no cached specifications
    # 1.0.3_nwbfile.nwb has cached "core" specification
    # 1.1.2_nwbfile.nwb has cached "core" and "hdmf-common" specifications

    # NOTE we use "coverage run -m pynwb.validate" instead of "python -m pynwb.validate"
    # so that we can both test pynwb.validate and compute code coverage from that test.
    # NOTE we also use "coverage run -p" which will generate a .coverage file with the
    # machine name, process id, and a random number appended to the filename to
    # simplify collecting and merging coverage data from multiple subprocesses. if "-p"
    # is not used, then each "coverage run" will overwrite the .coverage file from a
    # previous "coverage run".
    # NOTE the run_coverage.yml GitHub Action runs "python -m coverage combine" to
    # combine the individual coverage reports into one .coverage file.

    def test_validate_file_no_cache(self):
        """Test that validating a file with no cached spec against the core namespace succeeds."""
        result = subprocess.run(["coverage", "run", "-p", "-m", "pynwb.validate",
                                 "tests/back_compat/1.0.2_nwbfile.nwb"], capture_output=True)

        stderr_regex = re.compile(
            r".*UserWarning: No cached namespaces found in tests/back_compat/1\.0\.2_nwbfile\.nwb\s*"
            r"warnings.warn\(msg\)\s*"
            r"The file tests/back_compat/1\.0\.2_nwbfile\.nwb has no cached namespace information\. "
            r"Falling back to PyNWB namespace information\.\s*"
        )
        self.assertRegex(result.stderr.decode('utf-8'), stderr_regex)

        stdout_regex = re.compile(
            r"Validating tests/back_compat/1\.0\.2_nwbfile\.nwb against PyNWB namespace information using namespace "
            r"'core'\.\s* - no errors found\.\s*")
        self.assertRegex(result.stdout.decode('utf-8'), stdout_regex)

    def test_validate_file_no_cache_bad_ns(self):
        """Test that validating a file with no cached spec against a specified, unknown namespace fails."""
        result = subprocess.run(["coverage", "run", "-p", "-m", "pynwb.validate", "tests/back_compat/1.0.2_nwbfile.nwb",
                                 "--ns", "notfound"], capture_output=True)

        stderr_regex = re.compile(
            r".*UserWarning: No cached namespaces found in tests/back_compat/1\.0\.2_nwbfile\.nwb\s*"
            r"warnings.warn\(msg\)\s*"
            r"The file tests/back_compat/1\.0\.2_nwbfile\.nwb has no cached namespace information\. "
            r"Falling back to PyNWB namespace information\.\s*"
            r"The namespace 'notfound' could not be found in PyNWB namespace information as only "
            r"\['core'\] is present\.\s*"
        )
        self.assertRegex(result.stderr.decode('utf-8'), stderr_regex)

        self.assertEqual(result.stdout.decode('utf-8'), '')

    def test_validate_file_cached(self):
        """Test that validating a file with cached spec against its cached namespace succeeds."""
        result = subprocess.run(["coverage", "run", "-p", "-m", "pynwb.validate",
                                 "tests/back_compat/1.1.2_nwbfile.nwb"], capture_output=True)

        self.assertEqual(result.stderr.decode('utf-8'), '')

        stdout_regex = re.compile(
            r"Validating tests/back_compat/1\.1\.2_nwbfile\.nwb against cached namespace information using namespace "
            r"'core'\.\s* - no errors found\.\s*")
        self.assertRegex(result.stdout.decode('utf-8'), stdout_regex)

    def test_validate_file_cached_bad_ns(self):
        """Test that validating a file with cached spec against a specified, unknown namespace fails."""
        result = subprocess.run(["coverage", "run", "-p", "-m", "pynwb.validate",
                                 "tests/back_compat/1.1.2_nwbfile.nwb", "--ns", "notfound"], capture_output=True)

        stderr_regex = re.compile(
            r"The namespace 'notfound' could not be found in cached namespace information as only "
            r"\['core'\] is present\.\s*"
        )
        self.assertRegex(result.stderr.decode('utf-8'), stderr_regex)

        self.assertEqual(result.stdout.decode('utf-8'), '')

    def test_validate_file_cached_extension(self):
        """Test that validating a file with cached spec against the cached namespaces succeeds."""
        result = subprocess.run(["coverage", "run", "-p", "-m", "pynwb.validate",
                                 "tests/back_compat/2.1.0_nwbfile_with_extension.nwb"], capture_output=True)

        self.assertEqual(result.stderr.decode('utf-8'), '')

        stdout_regex = re.compile(
            r"Validating tests/back_compat/2\.1\.0_nwbfile_with_extension\.nwb against cached namespace information "
            r"using namespace 'ndx-testextension'\.\s* - no errors found\.\s*")
        self.assertRegex(result.stdout.decode('utf-8'), stdout_regex)

    def test_validate_file_cached_extension_pass_ns(self):
        """Test that validating a file with cached spec against the extension namespace succeeds."""
        result = subprocess.run(["coverage", "run", "-p", "-m", "pynwb.validate",
                                 "tests/back_compat/2.1.0_nwbfile_with_extension.nwb",
                                 "--ns", "ndx-testextension"], capture_output=True)

        self.assertEqual(result.stderr.decode('utf-8'), '')

        stdout_regex = re.compile(
            r"Validating tests/back_compat/2\.1\.0_nwbfile_with_extension\.nwb against cached namespace information "
            r"using namespace 'ndx-testextension'\.\s* - no errors found\.\s*")
        self.assertRegex(result.stdout.decode('utf-8'), stdout_regex)

    def test_validate_file_cached_core(self):
        """Test that validating a file with cached spec against the core namespace succeeds."""
        result = subprocess.run(["coverage", "run", "-p", "-m", "pynwb.validate",
                                 "tests/back_compat/2.1.0_nwbfile_with_extension.nwb",
                                 "--ns", "core"], capture_output=True)

        stdout_regex = re.compile(
            r"The namespace 'core' is included by the namespace 'ndx-testextension'. "
            r"Please validate against that namespace instead\.\s*"
        )
        self.assertRegex(result.stderr.decode('utf-8'), stdout_regex)

    def test_validate_file_cached_hdmf_common(self):
        """Test that validating a file with cached spec against the hdmf-common namespace fails."""
        result = subprocess.run(["coverage", "run", "-p", "-m", "pynwb.validate", "tests/back_compat/1.1.2_nwbfile.nwb",
                                 "--ns", "hdmf-common"], capture_output=True)

        stderr_regex = re.compile(
            r"The namespace 'hdmf-common' is included by the namespace 'core'\. Please validate against that "
            r"namespace instead\.\s*",
        )
        self.assertRegex(result.stderr.decode('utf-8'), stderr_regex)

    def test_validate_file_cached_ignore(self):
        """Test that validating a file with cached spec against the core namespace succeeds."""
        result = subprocess.run(["coverage", "run", "-p", "-m", "pynwb.validate", "tests/back_compat/1.1.2_nwbfile.nwb",
                                 "--no-cached-namespace"], capture_output=True)

        self.assertEqual(result.stderr.decode('utf-8'), '')

        stdout_regex = re.compile(
            r"Validating tests/back_compat/1\.1\.2_nwbfile\.nwb against PyNWB namespace information using namespace "
            r"'core'\.\s* - no errors found\.\s*")
        self.assertRegex(result.stdout.decode('utf-8'), stdout_regex)

    def test_validate_file_invalid(self):
        """Test that validating an invalid file outputs errors."""
        result = subprocess.run(
            [
                "coverage", "run", "-p", "-m", "pynwb.validate", "tests/back_compat/1.0.2_str_experimenter.nwb",
                "--no-cached-namespace"
            ],
            capture_output=True
        )

        stderr_regex = re.compile(
            r" - found the following errors:\s*"
            r"root/general/experimenter \(general/experimenter\): incorrect shape - expected an array of shape "
            r"'\[None\]', got non-array data 'one experimenter'\s*"
        )
        self.assertRegex(result.stderr.decode('utf-8'), stderr_regex)

        stdout_regex = re.compile(
            r"Validating tests/back_compat/1\.0\.2_str_experimenter\.nwb against PyNWB namespace information using "
            r"namespace 'core'\.\s*")
        self.assertRegex(result.stdout.decode('utf-8'), stdout_regex)

    def test_validate_file_list_namespaces_core(self):
        """Test listing namespaces from a file"""
        result = subprocess.run(
            [
                "coverage", "run", "-p", "-m", "pynwb.validate", "tests/back_compat/1.1.2_nwbfile.nwb",
                "--list-namespaces"
            ],
            capture_output=True
        )

        self.assertEqual(result.stderr.decode('utf-8'), '')

        stdout_regex = re.compile(r"core\s*")
        self.assertRegex(result.stdout.decode('utf-8'), stdout_regex)

    def test_validate_file_list_namespaces_extension(self):
        """Test listing namespaces from a file with an extension"""
        result = subprocess.run(
            [
                "coverage", "run", "-p", "-m", "pynwb.validate", "tests/back_compat/2.1.0_nwbfile_with_extension.nwb",
                "--list-namespaces"
            ],
            capture_output=True
        )

        self.assertEqual(result.stderr.decode('utf-8'), '')

        stdout_regex = re.compile(r"ndx-testextension\s*")
        self.assertRegex(result.stdout.decode('utf-8'), stdout_regex)


class TestValidateFunction(TestCase):

    # 1.0.2_nwbfile.nwb has no cached specifications
    # 1.0.3_nwbfile.nwb has cached "core" specification
    # 1.1.2_nwbfile.nwb has cached "core" and "hdmf-common" specificaitions

    def test_validate_io_no_cache(self):
        """Test that validating a file with no cached spec against the core namespace succeeds."""
        with NWBHDF5IO('tests/back_compat/1.0.2_nwbfile.nwb', 'r') as io:
            errors = validate(io)
            self.assertEqual(errors, [])

    def test_validate_io_no_cache_bad_ns(self):
        """Test that validating a file with no cached spec against a specified, unknown namespace fails."""
        with NWBHDF5IO('tests/back_compat/1.0.2_nwbfile.nwb', 'r') as io:
            with self.assertRaisesWith(KeyError, "\"'notfound' not a namespace\""):
                validate(io, 'notfound')

    def test_validate_io_cached(self):
        """Test that validating a file with cached spec against its cached namespace succeeds."""
        with NWBHDF5IO('tests/back_compat/1.1.2_nwbfile.nwb', 'r') as io:
            errors = validate(io)
            self.assertEqual(errors, [])

    def test_validate_io_cached_extension(self):
        """Test that validating a file with cached spec against its cached namespaces succeeds."""
        with NWBHDF5IO('tests/back_compat/2.1.0_nwbfile_with_extension.nwb', 'r', load_namespaces=True) as io:
            errors = validate(io)
            self.assertEqual(errors, [])

    def test_validate_io_cached_extension_pass_ns(self):
        """Test that validating a file with cached extension spec against the extension namespace succeeds."""
        with NWBHDF5IO('tests/back_compat/2.1.0_nwbfile_with_extension.nwb', 'r', load_namespaces=True) as io:
            errors = validate(io, 'ndx-testextension')
            self.assertEqual(errors, [])

    def test_validate_io_cached_core_with_io(self):
        """
        For back-compatability, test that validating a file with cached extension spec against the core
        namespace succeeds when using the `io` + `namespace` keywords.
        """
        with NWBHDF5IO(
            path='tests/back_compat/2.1.0_nwbfile_with_extension.nwb', mode='r', load_namespaces=True
        ) as io:
            results = validate(io=io, namespace="core")
            self.assertEqual(results, [])

    def test_validate_file_cached_extension(self):
        """
        Test that validating a file with cached extension spec against the core
        namespace raises an error with the new CLI-mimicing paths keyword.
        """
        nwbfile_path = "tests/back_compat/2.1.0_nwbfile_with_extension.nwb"
        with patch("sys.stderr", new=StringIO()) as fake_err:
            with patch("sys.stdout", new=StringIO()) as fake_out:
                results, status = validate(paths=[nwbfile_path], namespace="core", verbose=True)
                self.assertEqual(results, [])
                self.assertEqual(status, 1)
                self.assertEqual(
                    fake_err.getvalue(),
                    (
                        "The namespace 'core' is included by the namespace 'ndx-testextension'. "
                        "Please validate against that namespace instead.\n"
                    )
                )
                self.assertEqual(fake_out.getvalue(), "")

    def test_validate_file_cached_core(self):
        """
        Test that validating a file with cached core spec with verbose=False.
        """
        nwbfile_path = "tests/back_compat/1.1.2_nwbfile.nwb"
        with patch("sys.stderr", new=StringIO()) as fake_err:
            with patch("sys.stdout", new=StringIO()) as fake_out:
                results, status = validate(paths=[nwbfile_path], namespace="core")
                self.assertEqual(results, [])
                self.assertEqual(status, 0)
                self.assertEqual(fake_err.getvalue(), "")
                self.assertEqual(fake_out.getvalue(), "")

    def test_validate_file_cached_no_cache_bad_ns(self):
        """
        Test that validating a file with no cached namespace, a namespace that is not found, and verbose=False.
        """
        nwbfile_path = "tests/back_compat/1.0.2_nwbfile.nwb"
        with patch("sys.stderr", new=StringIO()) as fake_err:
            with patch("sys.stdout", new=StringIO()) as fake_out:
                results, status = validate(paths=[nwbfile_path], namespace="notfound")
                self.assertEqual(results, [])
                self.assertEqual(status, 1)
                stderr_regex = (
                    r"The namespace 'notfound' could not be found in PyNWB namespace information as only "
                    r"\['core'\] is present.\n"
                )
                self.assertRegex(fake_err.getvalue(), stderr_regex)
                self.assertEqual(fake_out.getvalue(), "")

    def test_validate_io_cached_bad_ns(self):
        """Test that validating a file with cached spec against a specified, unknown namespace fails."""
        with NWBHDF5IO('tests/back_compat/1.1.2_nwbfile.nwb', 'r') as io:
            with self.assertRaisesWith(KeyError, "\"'notfound' not a namespace\""):
                validate(io, 'notfound')

    def test_validate_io_cached_hdmf_common(self):
        """Test that validating a file with cached spec against the hdmf-common namespace fails."""
        with NWBHDF5IO('tests/back_compat/1.1.2_nwbfile.nwb', 'r') as io:
            # TODO this error should not be different from the error when using the validate script above
            msg = "builder must have data type defined with attribute 'data_type'"
            with self.assertRaisesWith(ValueError, msg):
                validate(io, 'hdmf-common')



from dandi.dandiapi import DandiAPIClient
import sys
import traceback

from pynwb import NWBHDF5IO
from pynwb.testing import TestCase


class TestReadNWBDandisets(TestCase):
    """Test reading NWB files from the DANDI Archive using ROS3."""

    def test_read_first_nwb_asset(self):
        """Test reading the first NWB asset from each dandiset that uses NWB."""
        client = DandiAPIClient()
        dandisets = client.get_dandisets()

        failed_reads = dict()
        for i, dandiset in enumerate(dandisets):
            dandiset_metadata = dandiset.get_raw_metadata()

            # skip any dandisets that do not use NWB
            if not any(
                data_standard["identifier"] == "RRID:SCR_015242"  # this is the RRID for NWB
                for data_standard in dandiset_metadata["assetsSummary"].get("dataStandard", [])
            ):
                continue

            dandiset_identifier = dandiset_metadata["identifier"]
            print("--------------")
            print(f"{i}: {dandiset_identifier}")

            # iterate through assets until we get an NWB file (it could be MP4)
            assets = dandiset.get_assets()
            first_asset = next(assets)
            while first_asset.path.split(".")[-1] != "nwb":
                first_asset = next(assets)
            if first_asset.path.split(".")[-1] != "nwb":
                print("No NWB files?!")
                continue

            s3_url = first_asset.get_content_url(follow_redirects=1, strip_query=True)

            try:
                with NWBHDF5IO(path=s3_url, load_namespaces=True, driver="ros3") as io:
                    io.read()
            except Exception as e:
                print(traceback.format_exc())
                failed_reads[dandiset] = e

        if failed_reads:
            print(failed_reads)
            sys.exit(1)



from ...device import Device

from .utils import name_generator


def mock_Device(
    name=None,
    description="description",
    manufacturer=None,
):
    return Device(
        name=name or name_generator("Device"),
        description=description,
        manufacturer=manufacturer,
    )



import numpy as np

from ...behavior import (
    PupilTracking,
    Position,
    SpatialSeries,
    CompassDirection,
)
from .utils import name_generator
from .base import mock_TimeSeries


def mock_SpatialSeries(
    name=None,
    data=None,
    reference_frame="lower left is (0, 0)",
    unit="meters",
    conversion=1.0,
    resolution=-1.0,
    timestamps=None,
    starting_time=None,
    rate=10.0,
    comments="no comments",
    description="no description",
    control=None,
    control_description=None,
):
    return SpatialSeries(
        name=name or name_generator("SpatialSeries"),
        data=data if data is not None else np.array([1, 2, 3, 4]),
        reference_frame=reference_frame,
        unit=unit,
        conversion=conversion,
        resolution=resolution,
        timestamps=timestamps,
        starting_time=starting_time,
        rate=rate,
        comments=comments,
        description=description,
        control=control,
        control_description=control_description,
    )


def mock_Position(
    name=None, spatial_series=None,
):
    return Position(name=name or name_generator("Position"), spatial_series=spatial_series or [mock_SpatialSeries()])


def mock_PupilTracking(
    name=None, time_series=None,
):
    return PupilTracking(name=name or name_generator("PupilTracking"), time_series=time_series or [mock_TimeSeries()])


def mock_CompassDirection(name=None, spatial_series=None):
    return CompassDirection(
        name=name or name_generator("CompassDirection"),
        spatial_series=spatial_series or [mock_SpatialSeries(unit="radians")],
    )



from warnings import warn
from collections.abc import Iterable
from typing import NamedTuple

import numpy as np

from hdmf.utils import docval, popargs_to_dict, get_docval, popargs
from hdmf.common import DynamicTable, VectorData
from hdmf.utils import get_data_shape

from . import register_class, CORE_NAMESPACE
from .core import NWBDataInterface, MultiContainerInterface, NWBData


@register_class('ProcessingModule', CORE_NAMESPACE)
class ProcessingModule(MultiContainerInterface):
    """ Processing module. This is a container for one or more containers
        that provide data at intermediate levels of analysis

        ProcessingModules should be created through calls to NWB.create_module().
        They should not be instantiated directly
    """

    __nwbfields__ = ('description',)

    __clsconf__ = {
            'attr': 'data_interfaces',
            'add': 'add',
            'type': (NWBDataInterface, DynamicTable),
            'get': 'get'
    }

    @docval({'name': 'name', 'type': str, 'doc': 'The name of this processing module'},
            {'name': 'description', 'type': str, 'doc': 'Description of this processing module'},
            {'name': 'data_interfaces', 'type': (list, tuple, dict),
             'doc': 'NWBDataInterfaces that belong to this ProcessingModule', 'default': None})
    def __init__(self, **kwargs):
        description, data_interfaces = popargs("description", "data_interfaces", kwargs)
        super().__init__(**kwargs)
        self.description = description
        self.data_interfaces = data_interfaces

    @property
    def containers(self):
        return self.data_interfaces

    @docval({'name': 'container', 'type': (NWBDataInterface, DynamicTable),
             'doc': 'the NWBDataInterface to add to this Module'})
    def add_container(self, **kwargs):
        '''
        Add an NWBContainer to this ProcessingModule
        '''
        warn(PendingDeprecationWarning('add_container will be replaced by add'))
        self.add(kwargs['container'])

    @docval({'name': 'container_name', 'type': str, 'doc': 'the name of the NWBContainer to retrieve'})
    def get_container(self, **kwargs):
        '''
        Retrieve an NWBContainer from this ProcessingModule
        '''
        warn(PendingDeprecationWarning('get_container will be replaced by get'))
        return self.get(kwargs['container_name'])

    @docval({'name': 'NWBDataInterface', 'type': (NWBDataInterface, DynamicTable),
             'doc': 'the NWBDataInterface to add to this Module'})
    def add_data_interface(self, **kwargs):
        warn(PendingDeprecationWarning('add_data_interface will be replaced by add'))
        self.add(kwargs['NWBDataInterface'])

    @docval({'name': 'data_interface_name', 'type': str, 'doc': 'the name of the NWBContainer to retrieve'})
    def get_data_interface(self, **kwargs):
        warn(PendingDeprecationWarning('get_data_interface will be replaced by get'))
        return self.get(kwargs['data_interface_name'])


@register_class('TimeSeries', CORE_NAMESPACE)
class TimeSeries(NWBDataInterface):
    """A generic base class for time series data"""
    __nwbfields__ = ("comments",
                     "description",
                     "data",
                     "resolution",
                     "conversion",
                     "offset",
                     "unit",
                     "timestamps",
                     "timestamps_unit",
                     "interval",
                     "starting_time",
                     "rate",
                     "starting_time_unit",
                     "control",
                     "control_description",
                     "continuity")

    __time_unit = "seconds"

    # values used when a TimeSeries is read and missing required fields
    DEFAULT_DATA = np.ndarray(shape=(0, ), dtype=np.uint8)
    DEFAULT_UNIT = 'unknown'

    DEFAULT_RESOLUTION = -1.0
    DEFAULT_CONVERSION = 1.0
    DEFAULT_OFFSET = 0.0

    @docval({'name': 'name', 'type': str, 'doc': 'The name of this TimeSeries dataset'},  # required
            {'name': 'data', 'type': ('array_data', 'data', 'TimeSeries'),
             'doc': ('The data values. The first dimension must be time. '
                     'Can also store binary data, e.g., image frames')},
            {'name': 'unit', 'type': str, 'doc': 'The base unit of measurement (should be SI unit)'},
            {'name': 'resolution', 'type': float,
             'doc': 'The smallest meaningful difference (in specified unit) between values in data',
             'default': DEFAULT_RESOLUTION},
            {'name': 'conversion', 'type': float,
             'doc': 'Scalar to multiply each element in data to convert it to the specified unit',
             'default': DEFAULT_CONVERSION},
            {
                'name': 'offset',
                'type': float,
                'doc': (
                    "Scalar to add to each element in the data scaled by 'conversion' to finish converting it to the "
                    "specified unit."
                    ),
                'default': DEFAULT_OFFSET
            },
            {'name': 'timestamps', 'type': ('array_data', 'data', 'TimeSeries'), 'shape': (None,),
             'doc': 'Timestamps for samples stored in data', 'default': None},
            {'name': 'starting_time', 'type': float, 'doc': 'The timestamp of the first sample', 'default': None},
            {'name': 'rate', 'type': float, 'doc': 'Sampling rate in Hz', 'default': None},

            {'name': 'comments', 'type': str, 'doc': 'Human-readable comments about this TimeSeries dataset',
             'default': 'no comments'},
            {'name': 'description', 'type': str, 'doc': 'Description of this TimeSeries dataset',
             'default': 'no description'},
            {'name': 'control', 'type': Iterable, 'doc': 'Numerical labels that apply to each element in data',
             'default': None},
            {'name': 'control_description', 'type': Iterable, 'doc': 'Description of each control value',
             'default': None},
            {'name': 'continuity', 'type': str, 'default': None, 'enum': ["continuous", "instantaneous", "step"],
             'doc': 'Optionally describe the continuity of the data. Can be "continuous", "instantaneous", or'
                    '"step". For example, a voltage trace would be "continuous", because samples are recorded from a '
                    'continuous process. An array of lick times would be "instantaneous", because the data represents '
                    'distinct moments in time. Times of image presentations would be  "step" because the picture '
                    'remains the same until the next time-point. This field is optional, but is useful in providing '
                    'information about the underlying data. It may inform the way this data is interpreted, the way it '
                    'is visualized, and what analysis methods are applicable.'})
    def __init__(self, **kwargs):
        """Create a TimeSeries object
        """
        keys_to_set = ("starting_time",
                       "rate",
                       "resolution",
                       "comments",
                       "description",
                       "conversion",
                       "offset",
                       "unit",
                       "control",
                       "control_description",
                       "continuity")
        args_to_set = popargs_to_dict(keys_to_set, kwargs)
        keys_to_process = ("data", "timestamps")  # these are properties and cannot be set with setattr
        args_to_process = popargs_to_dict(keys_to_process, kwargs)
        super().__init__(**kwargs)

        for key, val in args_to_set.items():
            setattr(self, key, val)

        data = args_to_process['data']
        self.fields['data'] = data
        if isinstance(data, TimeSeries):
            data.__add_link('data_link', self)

        timestamps = args_to_process['timestamps']
        if timestamps is not None:
            if self.rate is not None:
                raise ValueError('Specifying rate and timestamps is not supported.')
            if self.starting_time is not None:
                raise ValueError('Specifying starting_time and timestamps is not supported.')
            self.fields['timestamps'] = timestamps
            self.timestamps_unit = self.__time_unit
            self.interval = 1
            if isinstance(timestamps, TimeSeries):
                timestamps.__add_link('timestamp_link', self)
        elif self.rate is not None:
            if self.starting_time is None:  # override default if rate is provided but not starting time
                self.starting_time = 0.0
            self.starting_time_unit = self.__time_unit
        else:
            raise TypeError("either 'timestamps' or 'rate' must be specified")

        if not self._check_time_series_dimension():
            warn("%s '%s': Length of data does not match length of timestamps. Your data may be transposed. "
                 "Time should be on the 0th dimension" % (self.__class__.__name__, self.name))

    def _check_time_series_dimension(self):
        """Check that the 0th dimension of data equals the length of timestamps, when applicable.
        """
        if self.timestamps is None:
            return True

        data_shape = get_data_shape(data=self.fields["data"], strict_no_data_load=True)
        timestamps_shape = get_data_shape(data=self.fields["timestamps"], strict_no_data_load=True)

        # skip check if shape of data or timestamps cannot be computed
        if data_shape is None or timestamps_shape is None:
            return True

        # skip check if length of the first dimension is not known
        if data_shape[0] is None or timestamps_shape[0] is None:
            return True

        return data_shape[0] == timestamps_shape[0]

    @property
    def num_samples(self):
        ''' Tries to return the number of data samples. If this cannot be assessed, returns None.
        '''

        def unreadable_warning(attr):
            return (
                'The {} attribute on this TimeSeries (named: {}) has a __len__, '
                'but it cannot be read'.format(attr, self.name)
            )

        def no_len_warning(attr):
            return 'The {} attribute on this TimeSeries (named: {}) has no __len__'.format(attr, self.name)

        if hasattr(self.data, '__len__'):
            try:
                return len(self.data)  # for an ndarray this will return the first element of shape
            except TypeError:
                warn(unreadable_warning('data'), UserWarning)
        else:
            warn(no_len_warning('data'), UserWarning)

        # only get here if self.data has no __len__ or __len__ is unreadable
        if hasattr(self.timestamps, '__len__'):
            try:
                return len(self.timestamps)
            except TypeError:
                warn(unreadable_warning('timestamps'), UserWarning)
        elif self.rate is None and self.starting_time is None:
            warn(no_len_warning('timestamps'), UserWarning)

        return None

    @property
    def data(self):
        if isinstance(self.fields['data'], TimeSeries):
            return self.fields['data'].data
        else:
            return self.fields['data']

    @property
    def data_link(self):
        return self.__get_links('data_link')

    @property
    def timestamps(self):
        if 'timestamps' not in self.fields:
            return None
        if isinstance(self.fields['timestamps'], TimeSeries):
            return self.fields['timestamps'].timestamps
        else:
            return self.fields['timestamps']

    @property
    def timestamp_link(self):
        return self.__get_links('timestamp_link')

    def __get_links(self, links):
        ret = self.fields.get(links, list())
        ret = set(ret)
        return ret

    def __add_link(self, links_key, link):
        self.fields.setdefault(links_key, list()).append(link)

    @property
    def time_unit(self):
        return self.__time_unit


@register_class('Image', CORE_NAMESPACE)
class Image(NWBData):
    """
    Abstract image class. It is recommended to instead use pynwb.image.GrayscaleImage or pynwb.image.RGPImage where
    appropriate.
    """
    __nwbfields__ = ('data', 'resolution', 'description')

    @docval({'name': 'name', 'type': str, 'doc': 'The name of this image'},
            {'name': 'data', 'type': ('array_data', 'data'), 'doc': 'data of image. Dimensions: x, y [, r,g,b[,a]]',
             'shape': ((None, None), (None, None, 3), (None, None, 4))},
            {'name': 'resolution', 'type': float, 'doc': 'pixels / cm', 'default': None},
            {'name': 'description', 'type': str, 'doc': 'description of image', 'default': None})
    def __init__(self, **kwargs):
        args_to_set = popargs_to_dict(("resolution", "description"), kwargs)
        super().__init__(**kwargs)

        for key, val in args_to_set.items():
            setattr(self, key, val)


@register_class('ImageReferences', CORE_NAMESPACE)
class ImageReferences(NWBData):
    """
    Ordered dataset of references to Image objects.
    """
    __nwbfields__ = ('data', )

    @docval({'name': 'name', 'type': str, 'doc': 'The name of this ImageReferences object.'},
            {'name': 'data', 'type': 'array_data', 'doc': 'The images in order.'},)
    def __init__(self, **kwargs):
        # NOTE we do not use the docval shape validator here because it will recognize a list of P MxN images as
        # having shape (P, M, N)
        # check type and dimensionality
        for image in kwargs['data']:
            assert isinstance(image, Image), "Images used in ImageReferences must have type Image, not %s" % type(image)
        super().__init__(**kwargs)


@register_class('Images', CORE_NAMESPACE)
class Images(MultiContainerInterface):
    """An collection of images with an optional way to specify the order of the images
    using the "order_of_images" dataset. An order must be specified if the images are
    referenced by index, e.g., from an IndexSeries.
    """

    __nwbfields__ = ('description',
                     {'name': 'order_of_images', 'child': True, 'required_name': 'order_of_images'})

    __clsconf__ = {
        'attr': 'images',
        'add': 'add_image',
        'type': Image,
        'get': 'get_image',
        'create': 'create_image'
    }

    @docval({'name': 'name', 'type': str, 'doc': 'The name of this set of images'},
            {'name': 'images', 'type': 'array_data', 'doc': 'image objects', 'default': None},
            {'name': 'description', 'type': str, 'doc': 'description of images', 'default': 'no description'},
            {'name': 'order_of_images', 'type': ImageReferences,
             'doc': 'Ordered dataset of references to Image objects stored in the parent group.', 'default': None},)
    def __init__(self, **kwargs):

        args_to_set = popargs_to_dict(("description", "images", "order_of_images"), kwargs)
        super().__init__(**kwargs)
        for key, val in args_to_set.items():
            setattr(self, key, val)


class TimeSeriesReference(NamedTuple):
    """
    Class used to represent data values of a :py:class:`~pynwb.base.TimeSeriesReferenceVectorData`
    This is a ``typing.NamedTuple`` type with predefined tuple components
    :py:meth:`~pynwb.base.TimeSeriesReference.idx_start`, :py:meth:`~pynwb.base.TimeSeriesReference.count`, and
    :py:meth:`~pynwb.base.TimeSeriesReference.timeseries`.

    :cvar idx_start:
    :cvar count:
    :cvar timeseries:
    """
    idx_start: int
    """Start index in time for the timeseries"""

    count: int
    """Number of timesteps to be selected starting from :py:meth:`~pynwb.base.TimeSeriesReference.idx_start`"""

    timeseries: TimeSeries
    """The :py:class:`~pynwb.base.TimeSeries` object the TimeSeriesReference applies to"""

    def check_types(self):
        """
        Helper function to check correct types for :py:meth:`~pynwb.base.TimeSeriesReference.idx_start`,
        :py:meth:`~pynwb.base.TimeSeriesReference.count`, and :py:meth:`~pynwb.base.TimeSeriesReference.timeseries`.

        This function is usually used in the try/except block to check for `TypeError` raised by the function.

        See also :py:meth:`~pynwb.base.TimeSeriesReference.isvalid` to check both validity of types and the
        reference itself.

        :returns: True if successful. If unsuccessful `TypeError` is raised.

        :raises TypeError: If one of the fields does not match the expected type
        """
        if not isinstance(self.idx_start, (int, np.integer)):
            raise TypeError("idx_start must be an integer not %s" % str(type(self.idx_start)))
        if not isinstance(self.count, (int, np.integer)):
            raise TypeError("count must be an integer %s" % str(type(self.count)))
        if not isinstance(self.timeseries, TimeSeries):
            raise TypeError("timeseries must be of type TimeSeries. %s" % str(type(self.timeseries)))
        return True

    def isvalid(self):
        """
        Check whether the reference is valid. Setting :py:meth:`~pynwb.base.TimeSeriesReference.idx_start` and
        :py:meth:`~pynwb.base.TimeSeriesReference.count` to -1 is used to indicate invalid references. This is
        useful to allow for missing data in :py:class:`~pynwb.base.TimeSeriesReferenceVectorData`

        :returns: True if the selection is valid. Returns False if both
            :py:meth:`~pynwb.base.TimeSeriesReference.idx_start` and :py:meth:`~pynwb.base.TimeSeriesReference.count`
            are negative. Raises `IndexError` in case the indices are bad.

        :raises IndexError: If the combination of :py:meth:`~pynwb.base.TimeSeriesReference.idx_start` and
            :py:meth:`~pynwb.base.TimeSeriesReference.count` are not valid for the given timeseries.

        :raises TypeError: If one of the fields does not match the expected type
        """
        # Check types first
        self.check_types()
        # Check for none-type selection
        if self.idx_start < 0 and self.count < 0:
            return False
        num_samples = self.timeseries.num_samples
        if num_samples is not None:
            if self.idx_start >= num_samples or self.idx_start < 0:
                raise IndexError("'idx_start' %i out of range for timeseries '%s'" %
                                 (self.idx_start, self.timeseries.name))
            if self.count < 0:
                raise IndexError("'count' %i invalid. 'count' must be positive" % self.count)
            if (self.idx_start + self.count) > num_samples:
                raise IndexError("'idx_start + count' out of range for timeseries '%s'" % self.timeseries.name)
        return True

    @property
    def timestamps(self):
        """
        Get the floating point timestamp offsets in seconds from the timeseries that correspond to the array.
        These are either loaded directly from the :py:meth:`~pynwb.base.TimeSeriesReference.timeseries`
        timestamps or calculated from the starting time and sampling rate.


        :raises IndexError: If the combination of :py:meth:`~pynwb.base.TimeSeriesReference.idx_start` and
            :py:meth:`~pynwb.base.TimeSeriesReference.count` are not valid for the given timeseries.

        :raises TypeError: If one of the fields does not match the expected type

        :returns: Array with the timestamps.
        """
        # isvalid will be False only if both idx_start and count are negative. Otherwise well get errors or be True.
        if not self.isvalid():
            return None
        # load the data from the timestamps
        elif self.timeseries.timestamps is not None:
            return self.timeseries.timestamps[self.idx_start: (self.idx_start + self.count)]
        # construct the timestamps from the starting_time and rate
        else:
            start_time = self.timeseries.rate * self.idx_start + self.timeseries.starting_time
            return np.arange(0, self.count) * self.timeseries.rate + start_time

    @property
    def data(self):
        """
        Get the selected data values. This is a convenience function to slice data from the
        :py:meth:`~pynwb.base.TimeSeriesReference.timeseries` based on the given
        :py:meth:`~pynwb.base.TimeSeriesReference.idx_start` and
        :py:meth:`~pynwb.base.TimeSeriesReference.count`

        :raises IndexError: If the combination of :py:meth:`~pynwb.base.TimeSeriesReference.idx_start` and
            :py:meth:`~pynwb.base.TimeSeriesReference.count` are not valid for the given timeseries.

        :raises TypeError: If one of the fields does not match the expected type

        :returns: Result of ``self.timeseries.data[self.idx_start: (self.idx_start + self.count)]``. Returns
            None in case the reference is invalid (i.e., if both :py:meth:`~pynwb.base.TimeSeriesReference.idx_start`
            and :py:meth:`~pynwb.base.TimeSeriesReference.count` are negative.
        """
        # isvalid will be False only if both idx_start and count are negative. Otherwise well get errors or be True.
        if not self.isvalid():
            return None
        # load the data from the timeseries
        return self.timeseries.data[self.idx_start: (self.idx_start + self.count)]


@register_class('TimeSeriesReferenceVectorData', CORE_NAMESPACE)
class TimeSeriesReferenceVectorData(VectorData):
    """
    Column storing references to a TimeSeries (rows). For each TimeSeries this VectorData
    column stores the start_index and count to indicate the range in time to be selected
    as well as an object reference to the TimeSeries.

    **Representing missing values** In practice we sometimes need to be able to represent missing values,
    e.g., in the :py:class:`~pynwb.icephys.IntracellularRecordingsTable` we have
    :py:class:`~pynwb.base.TimeSeriesReferenceVectorData` to link to stimulus and
    response recordings, but a user can specify either only one of them or both. Since there is no
    ``None`` value for a complex types like ``(idx_start, count, TimeSeries)``, NWB defines
    ``None`` as ``(-1, -1, TimeSeries)`` for storage, i.e., if the ``idx_start`` (and ``count``) is negative
    then this indicates an invalid link (in practice both ``idx_start`` and ``count`` must always
    either both be positive or both be negative). When selecting data via the
    :py:meth:`~pynwb.base.TimeSeriesReferenceVectorData.get` or
    :py:meth:`~pynwb.base. TimeSeriesReferenceVectorData.__getitem__`
    functions, ``(-1, -1, TimeSeries)`` values are replaced by the corresponding
    :py:class:`~pynwb.base.TimeSeriesReferenceVectorData.TIME_SERIES_REFERENCE_NONE_TYPE` tuple
    to avoid exposing NWB storage internals to the user and simplifying the use of and checking
    for missing values. **NOTE:** We can still inspect the raw data values by looking at ``self.data``
    directly instead.

    :cvar TIME_SERIES_REFERENCE_TUPLE:
    :cvar TIME_SERIES_REFERENCE_NONE_TYPE:
    """

    TIME_SERIES_REFERENCE_TUPLE = TimeSeriesReference
    """Return type when calling :py:meth:`~pynwb.base.TimeSeriesReferenceVectorData.get` or
    :py:meth:`~pynwb.base. TimeSeriesReferenceVectorData.__getitem__`."""

    TIME_SERIES_REFERENCE_NONE_TYPE = TIME_SERIES_REFERENCE_TUPLE(None, None, None)
    """Tuple used to represent None values when calling :py:meth:`~pynwb.base.TimeSeriesReferenceVectorData.get` or
    :py:meth:`~pynwb.base. TimeSeriesReferenceVectorData.__getitem__`. See also
    :py:class:`~pynwb.base.TimeSeriesReferenceVectorData.TIME_SERIES_REFERENCE_TUPLE`"""

    @docval({'name': 'name', 'type': str, 'doc': 'the name of this VectorData', 'default': 'timeseries'},
            {'name': 'description', 'type': str, 'doc': 'a description for this column',
             'default': "Column storing references to a TimeSeries (rows). For each TimeSeries this "
                        "VectorData column stores the start_index and count to indicate the range in time "
                        "to be selected as well as an object reference to the TimeSeries."},
            *get_docval(VectorData.__init__, 'data'))
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # CAUTION: Define any logic specific for init in the self._init_internal function, not here!
        self._init_internal()

    def _init_internal(self):
        """
        Called from __init__ to perform initialization specific to this class. This is done
        here due to the :py:class:`~pynwb.io.epoch.TimeIntervalsMap` having to migrate legacy VectorData
        to TimeSeriesReferenceVectorData. In this way, if dedicated logic init logic needs
        to be added to this class then we have a place for it without having to also
        update :py:class:`~pynwb.io.epoch.TimeIntervalsMap` (which would likely get forgotten)
        """
        pass

    @docval({'name': 'val', 'type': (TIME_SERIES_REFERENCE_TUPLE, tuple),
             'doc': 'the value to add to this column. If this is a regular tuple then it '
                    'must be convertible to a TimeSeriesReference'})
    def add_row(self, **kwargs):
        """Append a data value to this column."""
        val = kwargs['val']
        if not isinstance(val, self.TIME_SERIES_REFERENCE_TUPLE):
            val = self.TIME_SERIES_REFERENCE_TUPLE(*val)
        val.check_types()
        super().append(val)

    @docval({'name': 'arg', 'type': (TIME_SERIES_REFERENCE_TUPLE, tuple),
             'doc': 'the value to append to this column. If this is a regular tuple then it '
                    'must be convertible to a TimeSeriesReference'})
    def append(self, **kwargs):
        """Append a data value to this column."""
        arg = kwargs['arg']
        if not isinstance(arg, self.TIME_SERIES_REFERENCE_TUPLE):
            arg = self.TIME_SERIES_REFERENCE_TUPLE(*arg)
        arg.check_types()
        super().append(arg)

    def get(self, key, **kwargs):
        """
        Retrieve elements from this object.

        The function uses :py:class:`~pynwb.base.TimeSeriesReferenceVectorData.TIME_SERIES_REFERENCE_TUPLE`
        to describe individual records in the dataset. This allows the code to avoid exposing internal
        details of the schema to the user and simplifies handling of missing values by explicitly
        representing missing values via
        :py:class:`~pynwb.base.TimeSeriesReferenceVectorData.TIME_SERIES_REFERENCE_NONE_TYPE`
        rather than the internal representation used for storage of ``(-1, -1, TimeSeries)``.

        :param key: Selection of the elements
        :param kwargs: Ignored

        :returns: :py:class:`~pynwb.base.TimeSeriesReferenceVectorData.TIME_SERIES_REFERENCE_TUPLE` if a single
                  element is being selected. Otherwise return a list of
                  :py:class:`~pynwb.base.TimeSeriesReferenceVectorData.TIME_SERIES_REFERENCE_TUPLE` objects.
                  Missing values are represented by
                  :py:class:`~pynwb.base.TimeSeriesReferenceVectorData.TIME_SERIES_REFERENCE_NONE_TYPE`
                  in which all values (i.e., idx_start, count, timeseries) are set to None.
        """
        vals = super().get(key)
        # we only selected one row.
        if isinstance(key, (int, np.integer)):
            # NOTE: If we never wrote the data to disk, then vals will be a single tuple.
            #       If the data is loaded from an h5py.Dataset then vals will be a single
            #       np.void object. I.e., an alternative check would be
            #       if isinstance(vals, tuple) or isinstance(vals, np.void):
            #          ...
            if vals[0] < 0 or vals[1] < 0:
                return self.TIME_SERIES_REFERENCE_NONE_TYPE
            else:
                return self.TIME_SERIES_REFERENCE_TUPLE(*vals)
        else:  # key selected multiple rows
            # When loading from HDF5 we get an np.ndarray otherwise we get list-of-list. This
            # makes the values consistent and transforms the data to use our namedtuple type
            re = [self.TIME_SERIES_REFERENCE_NONE_TYPE
                  if (v[0] < 0 or v[1] < 0) else self.TIME_SERIES_REFERENCE_TUPLE(*v)
                  for v in vals]
            return re



from hdmf.common.io.table import DynamicTableMap
from hdmf.common.io.alignedtable import AlignedDynamicTableMap

from .. import register_map
from .base import TimeSeriesMap
from pynwb.icephys import VoltageClampSeries, IntracellularRecordingsTable


@register_map(VoltageClampSeries)
class VoltageClampSeriesMap(TimeSeriesMap):

    def __init__(self, spec):
        super().__init__(spec)

        fields_with_unit = ('capacitance_fast',
                            'capacitance_slow',
                            'resistance_comp_bandwidth',
                            'resistance_comp_correction',
                            'resistance_comp_prediction',
                            'whole_cell_capacitance_comp',
                            'whole_cell_series_resistance_comp')
        for field in fields_with_unit:
            field_spec = self.spec.get_dataset(field)
            self.map_spec('%s__unit' % field, field_spec.get_attribute('unit'))


@register_map(IntracellularRecordingsTable)
class IntracellularRecordingsTableMap(AlignedDynamicTableMap):
    """
    Customize the mapping for AlignedDynamicTable
    """
    def __init__(self, spec):
        super().__init__(spec)

    @DynamicTableMap.object_attr('electrodes')
    def electrodes(self, container, manager):
        return container.category_tables.get('electrodes', None)

    @DynamicTableMap.object_attr('stimuli')
    def stimuli(self, container, manager):
        return container.category_tables.get('stimuli', None)

    @DynamicTableMap.object_attr('responses')
    def responses(self, container, manager):
        return container.category_tables.get('responses', None)



import warnings
from collections.abc import Iterable

import numpy as np

from hdmf.utils import (
    docval,
    getargs,
    popargs,
    popargs_to_dict,
    get_docval,
    get_data_shape,
)

from . import register_class, CORE_NAMESPACE
from .base import TimeSeries, Image, Images
from .device import Device


@register_class('ImageSeries', CORE_NAMESPACE)
class ImageSeries(TimeSeries):
    '''
    General image data that is common between acquisition and stimulus time series.
    The image data can be stored in the HDF5 file or it will be stored as an external image file.
    '''

    __nwbfields__ = ('dimension',
                     'external_file',
                     'starting_frame',
                     'format',
                     'device')

    # value used when an ImageSeries is read and missing data
    DEFAULT_DATA = np.ndarray(shape=(0, 0, 0), dtype=np.uint8)
    # TODO: copy new docs from 2.4 schema

    @docval(*get_docval(TimeSeries.__init__, 'name'),  # required
            {'name': 'data', 'type': ('array_data', 'data', TimeSeries), 'shape': ([None] * 3, [None] * 4),
             'doc': ('The data values. Can be 3D or 4D. The first dimension must be time (frame). The second and third '
                     'dimensions represent x and y. The optional fourth dimension represents z. Either data or '
                     'external_file must be specified (not None), but not both. If data is not specified, '
                     'data will be set to an empty 3D array.'),
             'default': None},
            {'name': 'unit', 'type': str,
             'doc': ('The unit of measurement of the image data, e.g., values between 0 and 255. Required when data '
                     'is specified. If unit (and data) are not specified, then unit will be set to "unknown".'),
             'default': None},
            {'name': 'format', 'type': str,
             'doc': 'Format of image. Three types: 1) Image format; tiff, png, jpg, etc. 2) external 3) raw.',
             'default': None},
            {'name': 'external_file', 'type': ('array_data', 'data'),
             'doc': 'Path or URL to one or more external file(s). Field only present if format=external. '
                    'Either external_file or data must be specified (not None), but not both.', 'default': None},
            {'name': 'starting_frame', 'type': Iterable,
             'doc': 'Each entry is a frame number that corresponds to the first frame of each file '
                    'listed in external_file within the full ImageSeries.', 'default': None},
            {'name': 'bits_per_pixel', 'type': int, 'doc': 'DEPRECATED: Number of bits per image pixel',
             'default': None},
            {'name': 'dimension', 'type': Iterable,
             'doc': 'Number of pixels on x, y, (and z) axes.', 'default': None},
            *get_docval(TimeSeries.__init__, 'resolution', 'conversion', 'timestamps', 'starting_time', 'rate',
                        'comments', 'description', 'control', 'control_description', 'offset'),
            {'name': 'device', 'type': Device,
             'doc': 'Device used to capture the images/video.', 'default': None},)
    def __init__(self, **kwargs):
        keys_to_set = ('bits_per_pixel', 'dimension', 'external_file', 'starting_frame', 'format', 'device')
        args_to_set = popargs_to_dict(keys_to_set, kwargs)
        name, data, unit = getargs('name', 'data', 'unit', kwargs)
        if data is not None and unit is None:
            raise ValueError("Must supply 'unit' argument when supplying 'data' to %s '%s'."
                             % (self.__class__.__name__, name))
        if args_to_set['external_file'] is None and data is None:
            raise ValueError("Must supply either external_file or data to %s '%s'."
                             % (self.__class__.__name__, name))

        # data and unit are required in TimeSeries, but allowed to be None here, so handle this specially
        if data is None:
            kwargs['data'] = ImageSeries.DEFAULT_DATA
        if unit is None:
            kwargs['unit'] = ImageSeries.DEFAULT_UNIT

        # If a single external_file is given then set starting_frame  to [0] for backward compatibility
        if (
            args_to_set["external_file"] is not None
            and args_to_set["starting_frame"] is None
        ):
            args_to_set["starting_frame"] = (
                [0] if len(args_to_set["external_file"]) == 1 else None
            )

        super().__init__(**kwargs)

        for key, val in args_to_set.items():
            setattr(self, key, val)

        if self._change_external_file_format():
            warnings.warn(
                "%s '%s': The value for 'format' has been changed to 'external'. "
                "Setting a default value for 'format' is deprecated and will be changed "
                "to raising a ValueError in the next major release."
                % (self.__class__.__name__, self.name),
                DeprecationWarning,
            )

        if not self._check_image_series_dimension():
            warnings.warn(
                "%s '%s': Length of data does not match length of timestamps. Your data may be transposed. "
                "Time should be on the 0th dimension"
                % (self.__class__.__name__, self.name)
            )

        self._error_on_new_warn_on_construct(
            error_msg=self._check_external_file_starting_frame_length()
        )
        self._error_on_new_warn_on_construct(
            error_msg=self._check_external_file_format()
        )
        self._error_on_new_warn_on_construct(error_msg=self._check_external_file_data())

    def _change_external_file_format(self):
        """
        Change the format to 'external' when external_file is specified.
        """
        if (
            get_data_shape(self.data)[0] == 0
            and self.external_file is not None
            and self.format is None
        ):
            self.format = "external"
            return True

        return False

    def _check_time_series_dimension(self):
        """Override _check_time_series_dimension to do nothing.
        The _check_image_series_dimension method will be called instead.
        """
        return True

    def _check_image_series_dimension(self):
        """Check that the 0th dimension of data equals the length of timestamps, when applicable.

        ImageSeries objects can have an external file instead of data stored. The external file cannot be
        queried for the number of frames it contains, so this check will return True when an external file
        is provided. Otherwise, this function calls the parent class' _check_time_series_dimension method.
        """
        if self.external_file is not None:
            return True
        return super()._check_time_series_dimension()

    def _check_external_file_starting_frame_length(self):
        """
        Check that the number of frame indices in 'starting_frame' matches
        the number of files in 'external_file'.
        """
        if self.external_file is None:
            return
        if get_data_shape(self.external_file) == get_data_shape(self.starting_frame):
            return

        return (
            "%s '%s': The number of frame indices in 'starting_frame' should have "
            "the same length as 'external_file'." % (self.__class__.__name__, self.name)
        )

    def _check_external_file_format(self):
        """
        Check that format is 'external' when external_file is specified.
        """
        if self.external_file is None:
            return
        if self.format == "external":
            return

        return "%s '%s': Format must be 'external' when external_file is specified." % (
            self.__class__.__name__,
            self.name,
        )

    def _check_external_file_data(self):
        """
        Check that data is an empty array when external_file is specified.
        """
        if self.external_file is None:
            return
        if get_data_shape(self.data)[0] == 0:
            return

        return (
            "%s '%s': Either external_file or data must be specified (not None), but not both."
            % (self.__class__.__name__, self.name)
        )

    @property
    def bits_per_pixel(self):
        return self.fields.get('bits_per_pixel')

    @bits_per_pixel.setter
    def bits_per_pixel(self, val):
        if val is not None:
            warnings.warn("bits_per_pixel is no longer used", DeprecationWarning)
            self.fields['bits_per_pixel'] = val


@register_class('IndexSeries', CORE_NAMESPACE)
class IndexSeries(TimeSeries):
    '''
    Stores indices to image frames stored in an ImageSeries. The purpose of the IndexSeries is to allow
    a static image stack to be stored somewhere, and the images in the stack to be referenced out-of-order.
    This can be for the display of individual images, or of movie segments (as a movie is simply a series of
    images). The data field stores the index of the frame in the referenced ImageSeries, and the timestamps
    array indicates when that image was displayed.
    '''

    __nwbfields__ = ("indexed_timeseries",)

    # # value used when an ImageSeries is read and missing data
    # DEFAULT_UNIT = 'N/A'

    @docval(
        *get_docval(TimeSeries.__init__, 'name'),  # required
        {
            'name': 'data',
            'type': ('array_data', 'data', TimeSeries),
            'shape': (None,),  # required
            'doc': 'The data values. Must be 1D, where the first dimension must be time (frame)',
        },
        *get_docval(TimeSeries.__init__, 'unit'),  # required
        {
            'name': 'indexed_timeseries', 'type': TimeSeries,  # required
            'doc': 'Link to TimeSeries containing images that are indexed.',
            'default': None,
        },
        {
            'name': 'indexed_images',
            'type': Images,  # required
            'doc': "Link to Images object containing an ordered set of images that are indexed. The Images object must "
                   "contain a 'ordered_images' dataset specifying the order of the images in the Images type.",
            'default': None
        },
        *get_docval(
            TimeSeries.__init__,
            'resolution',
            'conversion',
            'timestamps',
            'starting_time',
            'rate',
            'comments',
            'description',
            'control',
            'control_description',
            'offset',
        ),
    )
    def __init__(self, **kwargs):
        indexed_timeseries, indexed_images = popargs('indexed_timeseries', 'indexed_images', kwargs)
        if kwargs['unit'] and kwargs['unit'] != 'N/A':
            msg = ("The 'unit' field of IndexSeries is fixed to the value 'N/A' starting in NWB 2.5. Passing "
                   "a different value for 'unit' will raise an error in PyNWB 3.0.")
            warnings.warn(msg, PendingDeprecationWarning)
        if not indexed_timeseries and not indexed_images:
            msg = "Either indexed_timeseries or indexed_images must be provided when creating an IndexSeries."
            raise ValueError(msg)
        if indexed_timeseries:
            msg = ("The indexed_timeseries field of IndexSeries is discouraged and will be deprecated in "
                   "a future version of NWB. Use the indexed_images field instead.")
            warnings.warn(msg, PendingDeprecationWarning)
        kwargs['unit'] = 'N/A'  # fixed value starting in NWB 2.5
        super().__init__(**kwargs)
        self.indexed_timeseries = indexed_timeseries
        self.indexed_images = indexed_images
        if kwargs['conversion'] and kwargs['conversion'] != self.DEFAULT_CONVERSION:
            warnings.warn("The conversion attribute is not used by IndexSeries.")
        if kwargs['resolution'] and kwargs['resolution'] != self.DEFAULT_RESOLUTION:
            warnings.warn("The resolution attribute is not used by IndexSeries.")
        if kwargs['offset'] and kwargs['offset'] != self.DEFAULT_OFFSET:
            warnings.warn("The offset attribute is not used by IndexSeries.")


@register_class('ImageMaskSeries', CORE_NAMESPACE)
class ImageMaskSeries(ImageSeries):
    '''
    An alpha mask that is applied to a presented visual stimulus. The data[] array contains an array
    of mask values that are applied to the displayed image. Mask values are stored as RGBA. Mask
    can vary with time. The timestamps array indicates the starting time of a mask, and that mask
    pattern continues until it's explicitly changed.
    '''

    __nwbfields__ = ('masked_imageseries',)

    @docval(*get_docval(ImageSeries.__init__, 'name'),  # required
            {'name': 'masked_imageseries', 'type': ImageSeries,  # required
             'doc': 'Link to ImageSeries that mask is applied to.'},
            *get_docval(ImageSeries.__init__, 'data', 'unit', 'format', 'external_file', 'starting_frame',
                        'bits_per_pixel', 'dimension', 'resolution', 'conversion', 'timestamps', 'starting_time',
                        'rate', 'comments', 'description', 'control', 'control_description', 'offset'),
            {'name': 'device', 'type': Device,
             'doc': ('Device used to capture the mask data. This field will likely not be needed. '
                     'The device used to capture the masked ImageSeries data should be stored in the ImageSeries.'),
             'default': None},)
    def __init__(self, **kwargs):
        masked_imageseries = popargs('masked_imageseries', kwargs)
        super().__init__(**kwargs)
        self.masked_imageseries = masked_imageseries


@register_class('OpticalSeries', CORE_NAMESPACE)
class OpticalSeries(ImageSeries):
    '''
    Image data that is presented or recorded. A stimulus template movie will be stored only as an
    image. When the image is presented as stimulus, additional data is required, such as field of
    view (eg, how much of the visual field the image covers, or how what is the area of the target
    being imaged). If the OpticalSeries represents acquired imaging data, orientation is also
    important.
    '''

    __nwbfields__ = ('distance',
                     'field_of_view',
                     'orientation')

    @docval(*get_docval(ImageSeries.__init__, 'name'),  # required
            {'name': 'distance', 'type': float, 'doc': 'Distance from camera/monitor to target/eye.'},  # required
            {'name': 'field_of_view', 'type': ('array_data', 'data', 'TimeSeries'), 'shape': ((2, ), (3, )),  # required
             'doc': 'Width, height and depth of image, or imaged area (meters).'},
            {'name': 'orientation', 'type': str,  # required
             'doc': 'Description of image relative to some reference frame (e.g., which way is up). '
                    'Must also specify frame of reference.'},
            {'name': 'data', 'type': ('array_data', 'data'), 'shape': ([None] * 3, [None, None, None, 3]),
             'doc': ('Images presented to subject, either grayscale or RGB. May be 3D or 4D. The first dimension must '
                     'be time (frame). The second and third dimensions represent x and y. The optional fourth '
                     'dimension must be length 3 and represents the RGB value for color images. Either data or '
                     'external_file must be specified, but not both.'),
             'default': None},
            *get_docval(ImageSeries.__init__, 'unit', 'format', 'external_file', 'starting_frame', 'bits_per_pixel',
                        'dimension', 'resolution', 'conversion', 'timestamps', 'starting_time', 'rate', 'comments',
                        'description', 'control', 'control_description', 'device', 'offset'))
    def __init__(self, **kwargs):
        distance, field_of_view, orientation = popargs('distance', 'field_of_view', 'orientation', kwargs)
        super().__init__(**kwargs)
        self.distance = distance
        self.field_of_view = field_of_view
        self.orientation = orientation


@register_class('GrayscaleImage', CORE_NAMESPACE)
class GrayscaleImage(Image):

    @docval(*get_docval(Image.__init__, 'name'),
            {'name': 'data', 'type': ('array_data', 'data'),
             'doc': 'Data of grayscale image. Must be 2D where the dimensions represent x and y.',
             'shape': (None, None)},
            *get_docval(Image.__init__, 'resolution', 'description'))
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


@register_class('RGBImage', CORE_NAMESPACE)
class RGBImage(Image):

    @docval(*get_docval(Image.__init__, 'name'),
            {'name': 'data', 'type': ('array_data', 'data'),
             'doc': 'Data of color image. Must be 3D where the first and second dimensions represent x and y. '
                    'The third dimension has length 3 and represents the RGB value.',
             'shape': (None, None, 3)},
            *get_docval(Image.__init__, 'resolution', 'description'))
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


@register_class('RGBAImage', CORE_NAMESPACE)
class RGBAImage(Image):

    @docval(*get_docval(Image.__init__, 'name'),
            {'name': 'data', 'type': ('array_data', 'data'),
             'doc': 'Data of color image with transparency. Must be 3D where the first and second dimensions '
                    'represent x and y. The third dimension has length 4 and represents the RGBA value.',
             'shape': (None, None, 4)},
            *get_docval(Image.__init__, 'resolution', 'description'))
    def __init__(self, **kwargs):
        super().__init__(**kwargs)



from uuid import uuid4
from datetime import datetime
from dateutil.tz import tzlocal

from ...file import NWBFile, Subject
from .utils import name_generator


def mock_NWBFile(
    session_description='session_description',
    identifier=None,
    session_start_time=datetime(1970, 1, 1, tzinfo=tzlocal()),
    subject=None,
    **kwargs
):
    return NWBFile(
        session_description=session_description,
        identifier=identifier or str(uuid4()),
        session_start_time=session_start_time,
        subject=subject or mock_Subject(),
        **kwargs
    )


def mock_Subject(
    age="P50D",
    description="this is a mock mouse.",
    sex="F",
    subject_id=None,
):

    return Subject(
        age=age,
        description=description,
        sex=sex,
        subject_id=subject_id or name_generator("subject"),
    )



import numpy as np

from pynwb.base import (
    ProcessingModule,
    TimeSeries,
    Images,
    Image,
    TimeSeriesReferenceVectorData,
    TimeSeriesReference,
    ImageReferences
)
from pynwb.testing import TestCase
from hdmf.data_utils import DataChunkIterator
from hdmf.backends.hdf5 import H5DataIO


class TestProcessingModule(TestCase):
    def setUp(self):
        self.pm = ProcessingModule(
            name="test_procmod", description="a test processing module"
        )

    def _create_time_series(self):
        ts = TimeSeries(
            name="test_ts",
            data=[0, 1, 2, 3, 4, 5],
            unit="grams",
            timestamps=[0.0, 0.1, 0.2, 0.3, 0.4, 0.5],
        )
        return ts

    def test_init(self):
        """Test creating a ProcessingModule."""
        self.assertEqual(self.pm.name, "test_procmod")
        self.assertEqual(self.pm.description, "a test processing module")

    def test_add_data_interface(self):
        """Test adding a data interface to a ProcessingModule using add(...) and retrieving it."""
        ts = self._create_time_series()
        self.pm.add(ts)
        self.assertIn(ts.name, self.pm.containers)
        self.assertIs(ts, self.pm.containers[ts.name])

    def test_deprecated_add_data_interface(self):
        ts = self._create_time_series()
        with self.assertWarnsWith(
            PendingDeprecationWarning, "add_data_interface will be replaced by add"
        ):
            self.pm.add_data_interface(ts)
        self.assertIn(ts.name, self.pm.containers)
        self.assertIs(ts, self.pm.containers[ts.name])

    def test_deprecated_add_container(self):
        ts = self._create_time_series()
        with self.assertWarnsWith(
            PendingDeprecationWarning, "add_container will be replaced by add"
        ):
            self.pm.add_container(ts)
        self.assertIn(ts.name, self.pm.containers)
        self.assertIs(ts, self.pm.containers[ts.name])

    def test_get_data_interface(self):
        """Test adding a data interface to a ProcessingModule and retrieving it using get(...)."""
        ts = self._create_time_series()
        self.pm.add(ts)
        tmp = self.pm.get("test_ts")
        self.assertIs(tmp, ts)
        self.assertIs(self.pm["test_ts"], self.pm.get("test_ts"))

    def test_deprecated_get_data_interface(self):
        ts = self._create_time_series()
        self.pm.add(ts)
        with self.assertWarnsWith(
            PendingDeprecationWarning, "get_data_interface will be replaced by get"
        ):
            tmp = self.pm.get_data_interface("test_ts")
        self.assertIs(tmp, ts)

    def test_deprecated_get_container(self):
        ts = self._create_time_series()
        self.pm.add(ts)
        with self.assertWarnsWith(
            PendingDeprecationWarning, "get_container will be replaced by get"
        ):
            tmp = self.pm.get_container("test_ts")
        self.assertIs(tmp, ts)

    def test_getitem(self):
        """Test adding a data interface to a ProcessingModule and retrieving it using __getitem__(...)."""
        ts = self._create_time_series()
        self.pm.add(ts)
        tmp = self.pm["test_ts"]
        self.assertIs(tmp, ts)


class TestTimeSeries(TestCase):
    def test_init_no_parent(self):
        """Test creating an empty TimeSeries and that it has no parent."""
        ts = TimeSeries(name="test_ts", data=list(), unit="unit", timestamps=list())
        self.assertEqual(ts.name, "test_ts")
        self.assertIsNone(ts.parent)

    def test_init_datalink_set(self):
        """Test creating a TimeSeries and that data_link is an empty set."""
        ts = TimeSeries(name="test_ts", data=list(), unit="unit", timestamps=list())
        self.assertIsInstance(ts.data_link, set)
        self.assertEqual(len(ts.data_link), 0)

    def test_init_timestampslink_set(self):
        """Test creating a TimeSeries and that timestamps_link is an empty set."""
        ts = TimeSeries(name="test_ts", data=list(), unit="unit", timestamps=list())
        self.assertIsInstance(ts.timestamp_link, set)
        self.assertEqual(len(ts.timestamp_link), 0)

    def test_init_data_timestamps(self):
        data = [0, 1, 2, 3, 4]
        timestamps = [0.0, 0.1, 0.2, 0.3, 0.4]
        ts = TimeSeries(name="test_ts", data=data, unit="volts", timestamps=timestamps)
        self.assertIs(ts.data, data)
        self.assertIs(ts.timestamps, timestamps)
        self.assertEqual(ts.conversion, 1.0)
        self.assertEqual(ts.offset, 0.0)
        self.assertEqual(ts.resolution, -1.0)
        self.assertEqual(ts.unit, "volts")
        self.assertEqual(ts.interval, 1)
        self.assertEqual(ts.time_unit, "seconds")
        self.assertEqual(ts.num_samples, 5)
        self.assertIsNone(ts.continuity)
        self.assertIsNone(ts.rate)
        self.assertIsNone(ts.starting_time)

    def test_init_conversion_offset(self):
        data = [0, 1, 2, 3, 4]
        timestamps = [0.0, 0.1, 0.2, 0.3, 0.4]
        conversion = 2.1
        offset = 1.2
        ts = TimeSeries(
            name="test_ts",
            data=data,
            unit="volts",
            timestamps=timestamps,
            conversion=conversion,
            offset=offset,
        )
        self.assertIs(ts.data, data)
        self.assertEqual(ts.conversion, conversion)
        self.assertEqual(ts.offset, offset)

    def test_no_time(self):
        with self.assertRaisesWith(
            TypeError, "either 'timestamps' or 'rate' must be specified"
        ):
            TimeSeries(name="test_ts2", data=[10, 11, 12, 13, 14, 15], unit="grams")

    def test_no_starting_time(self):
        """Test that if no starting_time is given, 0.0 is assumed."""
        ts1 = TimeSeries(name="test_ts1", data=[1, 2, 3], unit="unit", rate=0.1)
        self.assertEqual(ts1.starting_time, 0.0)

    def test_init_rate(self):
        ts = TimeSeries(
            name="test_ts",
            data=list(),
            unit="volts",
            starting_time=1.0,
            rate=2.0,
        )
        self.assertEqual(ts.starting_time, 1.0)
        self.assertEqual(ts.starting_time_unit, "seconds")
        self.assertEqual(ts.rate, 2.0)
        self.assertEqual(ts.time_unit, "seconds")
        self.assertIsNone(ts.timestamps)

    def test_data_timeseries(self):
        """Test that setting a TimeSeries.data to another TimeSeries links the data correctly."""
        data = [0, 1, 2, 3]
        timestamps1 = [0.0, 0.1, 0.2, 0.3]
        timestamps2 = [1.0, 1.1, 1.2, 1.3]
        ts1 = TimeSeries(
            name="test_ts1", data=data, unit="grams", timestamps=timestamps1
        )
        ts2 = TimeSeries(
            name="test_ts2", data=ts1, unit="grams", timestamps=timestamps2
        )
        self.assertEqual(ts2.data, data)
        self.assertEqual(ts1.num_samples, ts2.num_samples)
        self.assertEqual(ts1.data_link, set([ts2]))

    def test_timestamps_timeseries(self):
        """Test that setting a TimeSeries.timestamps to another TimeSeries links the timestamps correctly."""
        data1 = [0, 1, 2, 3]
        data2 = [10, 11, 12, 13]
        timestamps = [0.0, 0.1, 0.2, 0.3]
        ts1 = TimeSeries(
            name="test_ts1", data=data1, unit="grams", timestamps=timestamps
        )
        ts2 = TimeSeries(name="test_ts2", data=data2, unit="grams", timestamps=ts1)
        self.assertEqual(ts2.timestamps, timestamps)
        self.assertEqual(ts1.timestamp_link, set([ts2]))

    def test_good_continuity_timeseries(self):
        ts = TimeSeries(
            name="test_ts1",
            data=[0, 1, 2, 3, 4, 5],
            unit="grams",
            timestamps=[0.0, 0.1, 0.2, 0.3, 0.4, 0.5],
            continuity="continuous",
        )
        self.assertEqual(ts.continuity, "continuous")

    def test_bad_continuity_timeseries(self):
        msg = (
            "TimeSeries.__init__: forbidden value for 'continuity' (got 'wrong', "
            "expected ['continuous', 'instantaneous', 'step'])"
        )
        with self.assertRaisesWith(ValueError, msg):
            TimeSeries(
                name="test_ts1",
                data=[0, 1, 2, 3, 4, 5],
                unit="grams",
                timestamps=[0.0, 0.1, 0.2, 0.3, 0.4, 0.5],
                continuity="wrong",
            )

    def _create_time_series_with_data(self, data):
        ts = TimeSeries(name="test_ts1", data=data, unit="grams", rate=0.1)
        return ts

    def test_dataio_list_data(self):
        length = 100
        data = list(range(length))
        ts = self._create_time_series_with_data(data)
        self.assertEqual(ts.num_samples, length)
        assert data == list(ts.data)

    def test_dataio_dci_data(self):
        def generator_factory():
            return (i for i in range(100))

        data = H5DataIO(DataChunkIterator(data=generator_factory()))
        ts = self._create_time_series_with_data(data)
        with self.assertWarnsWith(
            UserWarning,
            "The data attribute on this TimeSeries (named: test_ts1) has a "
            "__len__, but it cannot be read",
        ):
            self.assertIsNone(ts.num_samples)
        for xi, yi in zip(data, generator_factory()):
            assert np.allclose(xi, yi)

    def test_dci_data(self):
        def generator_factory():
            return (i for i in range(100))

        data = DataChunkIterator(data=generator_factory())
        ts = self._create_time_series_with_data(data)
        with self.assertWarnsWith(
            UserWarning,
            "The data attribute on this TimeSeries (named: test_ts1) has no __len__",
        ):
            self.assertIsNone(ts.num_samples)
        for xi, yi in zip(data, generator_factory()):
            assert np.allclose(xi, yi)

    def test_dci_data_arr(self):
        def generator_factory():
            return (np.array([i, i + 1]) for i in range(100))

        data = DataChunkIterator(data=generator_factory())
        ts = self._create_time_series_with_data(data)
        with self.assertWarnsWith(
            UserWarning,
            "The data attribute on this TimeSeries (named: test_ts1) has no __len__",
        ):
            self.assertIsNone(ts.num_samples)
        for xi, yi in zip(data, generator_factory()):
            assert np.allclose(xi, yi)

    def test_dataio_list_timestamps(self):
        length = 100
        data = list(range(length))
        ts = self._create_time_series_with_data(data)
        self.assertEqual(ts.num_samples, length)
        assert data == list(ts.data)

    def _create_time_series_with_timestamps(self, timestamps):
        # data has no __len__ for these tests
        def generator_factory():
            return (i for i in range(100))

        ts = TimeSeries(
            name="test_ts1",
            data=DataChunkIterator(data=generator_factory()),
            unit="grams",
            timestamps=timestamps,
        )
        return ts

    def test_dataio_dci_timestamps(self):
        def generator_factory():
            return (i for i in range(100))

        timestamps = H5DataIO(DataChunkIterator(data=generator_factory()))
        ts = self._create_time_series_with_timestamps(timestamps)
        with self.assertWarns(UserWarning) as record:
            self.assertIsNone(ts.num_samples)
        assert len(record.warnings) == 2
        assert record.warnings[0].message.args[0] == (
            "The data attribute on this TimeSeries (named: test_ts1) has no __len__"
        )
        assert record.warnings[1].message.args[0] == (
            "The timestamps attribute on this TimeSeries (named: test_ts1) has a "
            "__len__, but it cannot be read"
        )
        for xi, yi in zip(timestamps, generator_factory()):
            assert np.allclose(xi, yi)

    def test_dci_timestamps(self):
        def generator_factory():
            return (i for i in range(100))

        timestamps = DataChunkIterator(data=generator_factory())
        ts = self._create_time_series_with_timestamps(timestamps)
        with self.assertWarns(UserWarning) as record:
            self.assertIsNone(ts.num_samples)
        assert len(record.warnings) == 2
        assert record.warnings[0].message.args[0] == (
            "The data attribute on this TimeSeries (named: test_ts1) has no __len__"
        )
        assert record.warnings[1].message.args[0] == (
            "The timestamps attribute on this TimeSeries (named: test_ts1) has no __len__"
        )
        for xi, yi in zip(timestamps, generator_factory()):
            assert np.allclose(xi, yi)

    def test_dci_timestamps_arr(self):
        def generator_factory():
            return np.array(np.arange(100))

        timestamps = DataChunkIterator(data=generator_factory())
        ts = self._create_time_series_with_timestamps(timestamps)
        with self.assertWarns(UserWarning) as record:
            self.assertIsNone(ts.num_samples)
        assert len(record.warnings) == 2
        assert record.warnings[0].message.args[0] == (
            "The data attribute on this TimeSeries (named: test_ts1) has no __len__"
        )
        assert record.warnings[1].message.args[0] == (
            "The timestamps attribute on this TimeSeries (named: test_ts1) has no __len__"
        )
        for xi, yi in zip(timestamps, generator_factory()):
            assert np.allclose(xi, yi)

    def test_conflicting_time_args(self):
        with self.assertRaisesWith(
            ValueError, "Specifying rate and timestamps is not supported."
        ):
            TimeSeries(
                name="test_ts2",
                data=[10, 11, 12],
                unit="grams",
                rate=30.0,
                timestamps=[0.3, 0.4, 0.5],
            )
        with self.assertRaisesWith(
            ValueError, "Specifying starting_time and timestamps is not supported."
        ):
            TimeSeries(
                name="test_ts2",
                data=[10, 11, 12],
                unit="grams",
                starting_time=30.0,
                timestamps=[0.3, 0.4, 0.5],
            )

    def test_dimension_warning(self):
        msg = (
            "TimeSeries 'test_ts2': Length of data does not match length of timestamps. Your data may be "
            "transposed. Time should be on the 0th dimension"
        )
        with self.assertWarnsWith(UserWarning, msg):
            TimeSeries(
                name="test_ts2",
                data=[10, 11, 12],
                unit="grams",
                timestamps=[0.3, 0.4, 0.5, 0.6, 0.7, 0.8],
            )


class TestImage(TestCase):
    def test_init(self):
        im = Image(name="test_image", data=np.ones((10, 10)))
        assert im.name == "test_image"
        assert np.all(im.data == np.ones((10, 10)))


class TestImages(TestCase):

    def test_images(self):
        image1 = Image(name='test_image1', data=np.ones((10, 10)))
        image2 = Image(name='test_image2', data=np.ones((10, 10)))
        image_references = ImageReferences(name='order_of_images', data=[image2, image1])
        images = Images(name='images_name', images=[image1, image2], order_of_images=image_references)

        assert images.name == "images_name"
        assert images.images == dict(test_image1=image1, test_image2=image2)
        self.assertIs(images.order_of_images[0], image2)
        self.assertIs(images.order_of_images[1], image1)


class TestTimeSeriesReferenceVectorData(TestCase):
    def _create_time_series_with_rate(self):
        ts = TimeSeries(
            name="test",
            description="test",
            data=np.arange(10),
            unit="unit",
            starting_time=5.0,
            rate=0.1,
        )
        return ts

    def _create_time_series_with_timestamps(self):
        ts = TimeSeries(
            name="test",
            description="test",
            data=np.arange(10),
            unit="unit",
            timestamps=np.arange(10.0),
        )
        return ts

    def test_init(self):
        temp = TimeSeriesReferenceVectorData()
        self.assertEqual(temp.name, "timeseries")
        self.assertEqual(
            temp.description,
            "Column storing references to a TimeSeries (rows). For each TimeSeries this "
            "VectorData column stores the start_index and count to indicate the range in time "
            "to be selected as well as an object reference to the TimeSeries.",
        )
        self.assertListEqual(temp.data, [])
        temp = TimeSeriesReferenceVectorData(name="test", description="test")
        self.assertEqual(temp.name, "test")
        self.assertEqual(temp.description, "test")

    def test_get_empty(self):
        """Get data from an empty TimeSeriesReferenceVectorData"""
        temp = TimeSeriesReferenceVectorData()
        self.assertListEqual(temp[:], [])
        with self.assertRaises(IndexError):
            temp[0]

    def test_append_get_length1_valid_data(self):
        """Get data from a TimeSeriesReferenceVectorData with one element and valid data"""
        temp = TimeSeriesReferenceVectorData()
        value = TimeSeriesReference(0, 5, self._create_time_series_with_rate())
        temp.append(value)
        self.assertTupleEqual(temp[0], value)
        self.assertListEqual(
            temp[:],
            [
                TimeSeriesReferenceVectorData.TIME_SERIES_REFERENCE_TUPLE(*value),
            ],
        )

    def test_add_row_get_length1_valid_data(self):
        """Get data from a TimeSeriesReferenceVectorData with one element and valid data"""
        temp = TimeSeriesReferenceVectorData()
        value = TimeSeriesReference(0, 5, self._create_time_series_with_rate())
        temp.add_row(value)
        self.assertTupleEqual(temp[0], value)
        self.assertListEqual(
            temp[:],
            [
                TimeSeriesReferenceVectorData.TIME_SERIES_REFERENCE_TUPLE(*value),
            ],
        )

    def test_get_length1_invalid_data(self):
        """Get data from a TimeSeriesReferenceVectorData with one element and invalid data"""
        temp = TimeSeriesReferenceVectorData()
        value = TimeSeriesReference(-1, -1, self._create_time_series_with_rate())
        temp.append(value)
        # test index slicing
        re = temp[0]
        self.assertTrue(
            isinstance(re, TimeSeriesReferenceVectorData.TIME_SERIES_REFERENCE_TUPLE)
        )
        self.assertTupleEqual(
            re, TimeSeriesReferenceVectorData.TIME_SERIES_REFERENCE_NONE_TYPE
        )
        # test array slicing and list slicing
        selection = [
            slice(None),
            [
                0,
            ],
        ]
        for s in selection:
            re = temp[s]
            self.assertTrue(isinstance(re, list))
            self.assertTrue(len(re), 1)
            self.assertTrue(
                isinstance(
                    re[0], TimeSeriesReferenceVectorData.TIME_SERIES_REFERENCE_TUPLE
                )
            )
            self.assertTupleEqual(
                re[0], TimeSeriesReferenceVectorData.TIME_SERIES_REFERENCE_NONE_TYPE
            )

    def test_get_length5_valid_data(self):
        """Get data from a TimeSeriesReferenceVectorData with 5 elements"""
        temp = TimeSeriesReferenceVectorData()
        num_values = 5
        values = [
            TimeSeriesReference(0, 5, self._create_time_series_with_rate())
            for i in range(num_values)
        ]
        for v in values:
            temp.append(v)
        # Test single element selection
        for i in range(num_values):
            # test index slicing
            re = temp[i]
            self.assertTupleEqual(re, values[i])
            # test slicing
            re = temp[i : i + 1]
            self.assertTupleEqual(
                re[0],
                TimeSeriesReferenceVectorData.TIME_SERIES_REFERENCE_TUPLE(*values[i]),
            )
        # Test multi element selection
        re = temp[0:2]
        self.assertTupleEqual(
            re[0], TimeSeriesReferenceVectorData.TIME_SERIES_REFERENCE_TUPLE(*values[0])
        )
        self.assertTupleEqual(
            re[1], TimeSeriesReferenceVectorData.TIME_SERIES_REFERENCE_TUPLE(*values[1])
        )

    def test_get_length5_with_invalid_data(self):
        """Get data from a TimeSeriesReferenceVectorData with 5 elements"""
        temp = TimeSeriesReferenceVectorData()
        num_values = 5
        values = [
            TimeSeriesReference(0, 5, self._create_time_series_with_rate())
            for i in range(num_values - 2)
        ]
        values = (
            [
                TimeSeriesReference(-1, -1, self._create_time_series_with_rate()),
            ]
            + values
            + [
                TimeSeriesReference(-1, -1, self._create_time_series_with_rate()),
            ]
        )
        for v in values:
            temp.append(v)
        # Test single element selection
        for i in range(num_values):
            # test index slicing
            re = temp[i]
            if i in [0, 4]:
                self.assertTrue(
                    isinstance(
                        re, TimeSeriesReferenceVectorData.TIME_SERIES_REFERENCE_TUPLE
                    )
                )
                self.assertTupleEqual(
                    re, TimeSeriesReferenceVectorData.TIME_SERIES_REFERENCE_NONE_TYPE
                )
            else:
                self.assertTupleEqual(re, values[i])
            # test slicing
            re = temp[i : i + 1]
            if i in [0, 4]:
                self.assertTrue(
                    isinstance(
                        re[0], TimeSeriesReferenceVectorData.TIME_SERIES_REFERENCE_TUPLE
                    )
                )
                self.assertTupleEqual(
                    re[0], TimeSeriesReferenceVectorData.TIME_SERIES_REFERENCE_NONE_TYPE
                )
            else:
                self.assertTupleEqual(
                    re[0],
                    TimeSeriesReferenceVectorData.TIME_SERIES_REFERENCE_TUPLE(
                        *values[i]
                    ),
                )
        # Test multi element selection
        re = temp[0:2]
        self.assertTupleEqual(
            re[0], TimeSeriesReferenceVectorData.TIME_SERIES_REFERENCE_NONE_TYPE
        )
        self.assertTupleEqual(
            re[1], TimeSeriesReferenceVectorData.TIME_SERIES_REFERENCE_TUPLE(*values[1])
        )

    def test_add_row(self):
        v = TimeSeriesReferenceVectorData(name='a', description='a')
        val = TimeSeriesReference(0, 5, TimeSeries(name='test', description='test',
                                                   data=np.arange(10), unit='unit', starting_time=5.0, rate=0.1))
        v.add_row(val)
        self.assertTupleEqual(v[0], val)

    def test_add_row_with_plain_tuple(self):
        v = TimeSeriesReferenceVectorData(name='a', description='a')
        val = (0, 5, TimeSeries(name='test', description='test',
                                data=np.arange(10), unit='unit', starting_time=5.0, rate=0.1))
        v.add_row(val)
        self.assertTupleEqual(v[0], val)

    def test_add_row_with_bad_tuple(self):
        v = TimeSeriesReferenceVectorData(name='a', description='a')
        val = (0.0, 5, TimeSeries(name='test', description='test',
                                  data=np.arange(10), unit='unit', starting_time=5.0, rate=0.1))
        with self.assertRaisesWith(TypeError, "idx_start must be an integer not <class 'float'>"):
            v.add_row(val)

    def test_add_row_restricted_type(self):
        v = TimeSeriesReferenceVectorData(name="a", description="a")
        with self.assertRaisesWith(
            TypeError,
            "TimeSeriesReferenceVectorData.add_row: incorrect type for "
            "'val' (got 'int', expected 'TimeSeriesReference or tuple')",
        ):
            v.add_row(1)

    def test_append(self):
        v = TimeSeriesReferenceVectorData(name='a', description='a')
        val = TimeSeriesReference(0, 5, TimeSeries(name='test', description='test',
                                                   data=np.arange(10), unit='unit', starting_time=5.0, rate=0.1))
        v.append(val)
        self.assertTupleEqual(v[0], val)

    def test_append_with_plain_tuple(self):
        v = TimeSeriesReferenceVectorData(name='a', description='a')
        val = (0, 5, TimeSeries(name='test', description='test',
                                data=np.arange(10), unit='unit', starting_time=5.0, rate=0.1))
        v.append(val)
        self.assertTupleEqual(v[0], val)

    def test_append_with_bad_tuple(self):
        v = TimeSeriesReferenceVectorData(name='a', description='a')
        val = (0.0, 5, TimeSeries(name='test', description='test',
                                  data=np.arange(10), unit='unit', starting_time=5.0, rate=0.1))
        with self.assertRaisesWith(TypeError, "idx_start must be an integer not <class 'float'>"):
            v.append(val)

    def test_append_restricted_type(self):
        v = TimeSeriesReferenceVectorData(name="a", description="a")
        with self.assertRaisesWith(
            TypeError,
            "TimeSeriesReferenceVectorData.append: incorrect type for "
            "'arg' (got 'float', expected 'TimeSeriesReference or tuple')",
        ):
            v.append(2.0)


class TestTimeSeriesReference(TestCase):
    def _create_time_series_with_rate(self):
        ts = TimeSeries(
            name="test",
            description="test",
            data=np.arange(10),
            unit="unit",
            starting_time=5.0,
            rate=0.1,
        )
        return ts

    def _create_time_series_with_timestamps(self):
        ts = TimeSeries(
            name="test",
            description="test",
            data=np.arange(10),
            unit="unit",
            timestamps=np.arange(10.0),
        )
        return ts

    def test_check_types(self):
        # invalid selection but with correct types
        tsr = TimeSeriesReference(-1, -1, self._create_time_series_with_rate())
        self.assertTrue(tsr.check_types())
        # invalid types, use float instead of int for both idx_start and count
        tsr = TimeSeriesReference(1.0, 5.0, self._create_time_series_with_rate())
        with self.assertRaisesWith(
            TypeError, "idx_start must be an integer not <class 'float'>"
        ):
            tsr.check_types()
        # invalid types, use float instead of int for idx_start only
        tsr = TimeSeriesReference(1.0, 5, self._create_time_series_with_rate())
        with self.assertRaisesWith(
            TypeError, "idx_start must be an integer not <class 'float'>"
        ):
            tsr.check_types()
        # invalid types, use float instead of int for count only
        tsr = TimeSeriesReference(1, 5.0, self._create_time_series_with_rate())
        with self.assertRaisesWith(
            TypeError, "count must be an integer <class 'float'>"
        ):
            tsr.check_types()
        # invalid type for TimeSeries but valid idx_start and count
        tsr = TimeSeriesReference(1, 5, None)
        with self.assertRaisesWith(
            TypeError, "timeseries must be of type TimeSeries. <class 'NoneType'>"
        ):
            tsr.check_types()

    def test_is_invalid(self):
        tsr = TimeSeriesReference(-1, -1, self._create_time_series_with_rate())
        self.assertFalse(tsr.isvalid())

    def test_is_valid(self):
        tsr = TimeSeriesReference(0, 10, self._create_time_series_with_rate())
        self.assertTrue(tsr.isvalid())

    def test_is_valid_bad_index(self):
        # Error: negative start_index but positive count
        tsr = TimeSeriesReference(-1, 10, self._create_time_series_with_rate())
        with self.assertRaisesWith(
            IndexError, "'idx_start' -1 out of range for timeseries 'test'"
        ):
            tsr.isvalid()
        # Error: start_index too large
        tsr = TimeSeriesReference(10, 0, self._create_time_series_with_rate())
        with self.assertRaisesWith(
            IndexError, "'idx_start' 10 out of range for timeseries 'test'"
        ):
            tsr.isvalid()
        # Error: positive start_index but negative count
        tsr = TimeSeriesReference(0, -3, self._create_time_series_with_rate())
        with self.assertRaisesWith(
            IndexError, "'count' -3 invalid. 'count' must be positive"
        ):
            tsr.isvalid()
        # Error:  start_index + count too large
        tsr = TimeSeriesReference(3, 10, self._create_time_series_with_rate())
        with self.assertRaisesWith(
            IndexError, "'idx_start + count' out of range for timeseries 'test'"
        ):
            tsr.isvalid()

    def test_is_valid_no_num_samples(self):
        def generator_factory():
            return (i for i in range(100))

        data = DataChunkIterator(data=generator_factory())
        ts = TimeSeries(name="test_ts1", data=data, unit="grams", rate=0.1)
        tsr = TimeSeriesReference(0, 10, ts)
        with self.assertWarnsWith(
            UserWarning,
            "The data attribute on this TimeSeries (named: test_ts1) has no __len__",
        ):
            self.assertTrue(tsr.isvalid())

    def test_timestamps_property(self):
        # Timestamps from starting_time and rate
        tsr = TimeSeriesReference(5, 4, self._create_time_series_with_rate())
        np.testing.assert_array_equal(tsr.timestamps, np.array([5.5, 5.6, 5.7, 5.8]))
        # Timestamps from timestamps directly
        tsr = TimeSeriesReference(5, 4, self._create_time_series_with_timestamps())
        np.testing.assert_array_equal(tsr.timestamps, np.array([5.0, 6.0, 7.0, 8.0]))

    def test_timestamps_property_invalid_reference(self):
        # Timestamps from starting_time and rate
        tsr = TimeSeriesReference(-1, -1, self._create_time_series_with_rate())
        self.assertIsNone(tsr.timestamps)

    def test_timestamps_property_bad_reference(self):
        tsr = TimeSeriesReference(0, 12, self._create_time_series_with_timestamps())
        with self.assertRaisesWith(
            IndexError, "'idx_start + count' out of range for timeseries 'test'"
        ):
            tsr.timestamps
        tsr = TimeSeriesReference(0, 12, self._create_time_series_with_rate())
        with self.assertRaisesWith(
            IndexError, "'idx_start + count' out of range for timeseries 'test'"
        ):
            tsr.timestamps

    def test_data_property(self):
        tsr = TimeSeriesReference(5, 4, self._create_time_series_with_rate())
        np.testing.assert_array_equal(tsr.data, np.array([5.0, 6.0, 7.0, 8.0]))

    def test_data_property_invalid_reference(self):
        tsr = TimeSeriesReference(-1, -1, self._create_time_series_with_rate())
        self.assertIsNone(tsr.data)

    def test_data_property_bad_reference(self):
        tsr = TimeSeriesReference(0, 12, self._create_time_series_with_rate())
        with self.assertRaisesWith(
            IndexError, "'idx_start + count' out of range for timeseries 'test'"
        ):
            tsr.data



import numpy as np

from pynwb.icephys import (
    IntracellularElectrode,
    VoltageClampSeries,
    VoltageClampStimulusSeries,
    IntracellularRecordingsTable,
    CurrentClampSeries,
    CurrentClampStimulusSeries,
    IZeroClampSeries,
)

from .utils import name_generator
from .device import mock_Device


def mock_IntracellularElectrode(
    name=None, description="description", device=None,
):
    return IntracellularElectrode(
        name=name or name_generator("IntracellularElectrode"),
        description=description,
        device=device or mock_Device(),
    )


def mock_VoltageClampStimulusSeries(
    name=None,
    data=None,
    rate=100_000.,
    electrode=None,
    gain=0.02,
    timestamps=None,
    starting_time=None,
):
    return VoltageClampStimulusSeries(
        name=name or name_generator("VoltageClampStimulusSeries"),
        data=data or np.ones((30,)),
        rate=None if timestamps else rate,
        electrode=electrode or mock_IntracellularElectrode(),
        gain=gain,
        timestamps=timestamps,
        starting_time=starting_time,
    )


def mock_VoltageClampSeries(
    name=None,
    data=None,
    conversion=1.0,
    resolution=np.nan,
    starting_time=None,
    rate=100_000.0,
    electrode=None,
    gain=0.02,
    capacitance_slow=100e-12,
    resistance_comp_correction=70.0,
):
    return VoltageClampSeries(
        name=name if name is not None else name_generator("VoltageClampSeries"),
        data=data if data is not None else np.ones((30,)),
        conversion=conversion,
        resolution=resolution,
        starting_time=starting_time,
        rate=rate,
        electrode=electrode or mock_IntracellularElectrode(),
        gain=gain,
        capacitance_slow=capacitance_slow,
        resistance_comp_correction=resistance_comp_correction,
    )


def mock_CurrentClampSeries(
    name=None,
    data=None,
    electrode=None,
    gain=0.02,
    stimulus_description="N/A",
    bias_current=None,
    bridge_balance=None,
    capacitance_compensation=None,
    resolution=-1.0,
    conversion=1.0,
    timestamps=None,
    starting_time=None,
    rate=100_000.,
    comments="no comments",
    description="no description",
    control=None,
    control_description=None,
    sweep_number=None,
    offset=0.0,
    unit="volts",
):
    return CurrentClampSeries(
        name=name if name is not None else name_generator("CurrentClampSeries"),
        data=data if data is not None else np.ones((30,)),
        electrode=electrode or mock_IntracellularElectrode(),
        gain=gain,
        stimulus_description=stimulus_description,
        bias_current=bias_current,
        bridge_balance=bridge_balance,
        capacitance_compensation=capacitance_compensation,
        resolution=resolution,
        conversion=conversion,
        timestamps=timestamps,
        starting_time=starting_time,
        rate=rate,
        comments=comments,
        description=description,
        control=control,
        control_description=control_description,
        sweep_number=sweep_number,
        offset=offset,
        unit=unit,
    )


def mock_CurrentClampStimulusSeries(
    name=None,
    data=None,
    electrode=None,
    gain=0.02,
    stimulus_description="N/A",
    resolution=-1.0,
    conversion=1.0,
    timestamps=None,
    starting_time=None,
    rate=100_000.,
    comments="no comments",
    description="no description",
    control=None,
    control_description=None,
    sweep_number=None,
    offset=0.0,
    unit="amperes",
):
    return CurrentClampStimulusSeries(
        name=name or name_generator("CurrentClampStimulusSeries"),
        data=data if data is not None else np.ones((30,)),
        electrode=electrode or mock_IntracellularElectrode(),
        gain=gain,
        stimulus_description=stimulus_description,
        resolution=resolution,
        conversion=conversion,
        timestamps=timestamps,
        starting_time=starting_time,
        rate=rate,
        comments=comments,
        description=description,
        control=control,
        control_description=control_description,
        sweep_number=sweep_number,
        offset=offset,
        unit=unit,
    )


def mock_IZeroClampSeries(
    name=None,
    data=None,
    electrode=None,
    gain=.02,
    stimulus_description="N/A",
    resolution=-1.0,
    conversion=1.0,
    timestamps=None,
    starting_time=None,
    rate=100_000.,
    comments="no comments",
    description="no description",
    control=None,
    control_description=None,
    sweep_number=None,
    offset=0.0,
    unit="volts",
):
    return IZeroClampSeries(
        name=name or name_generator("IZeroClampSeries"),
        data=data if data is not None else np.ones((30,)),
        electrode=electrode or mock_IntracellularElectrode(),
        gain=gain,
        stimulus_description=stimulus_description,
        resolution=resolution,
        conversion=conversion,
        timestamps=timestamps,
        starting_time=starting_time,
        rate=rate,
        comments=comments,
        description=description,
        control=control,
        control_description=control_description,
        sweep_number=sweep_number,
        offset=offset,
        unit=unit,
    )


def mock_IntracellularRecordingsTable(n_rows=5):
    irt = IntracellularRecordingsTable()
    for _ in range(n_rows):
        electrode = mock_IntracellularElectrode()
        irt.add_recording(
            electrode=electrode,
            stimulus=mock_VoltageClampStimulusSeries(electrode=electrode),
            response=mock_VoltageClampSeries(electrode=electrode),
        )



import numpy as np

from hdmf.common.table import DynamicTableRegion

from ...file import ElectrodeTable
from ...ecephys import ElectricalSeries, ElectrodeGroup, SpikeEventSeries
from .device import mock_Device
from .utils import name_generator


def mock_ElectrodeGroup(
    name=None,
    description="description",
    location="location",
    device=None,
    position=None,
):
    return ElectrodeGroup(
        name=name or name_generator("ElectrodeGroup"),
        description=description,
        location=location,
        device=device or mock_Device(),
        position=position,
    )


def mock_ElectrodeTable(n_rows=5, group=None):
    table = ElectrodeTable()
    group = group if group is not None else mock_ElectrodeGroup()
    for i in range(n_rows):
        table.add_row(
            location="CA1",
            group=group,
            group_name=group.name,
        )
    return table


def mock_electrodes(n_electrodes=5, table=mock_ElectrodeTable(n_rows=5)):
    return DynamicTableRegion(
        name="electrodes",
        data=list(range(n_electrodes)),
        description="the first and third electrodes",
        table=table,
    )


def mock_ElectricalSeries(
    name=None,
    description="description",
    data=None,
    rate=30000.0,
    timestamps=None,
    electrodes=None,
    filtering="filtering",
):
    return ElectricalSeries(
        name=name or name_generator("ElectricalSeries"),
        description=description,
        data=data if data is not None else np.ones((10, 5)),
        rate=rate,
        timestamps=timestamps,
        electrodes=electrodes or mock_electrodes(),
        filtering=filtering,
    )


def mock_SpikeEventSeries(
    name=None,
    description="description",
    data=None,
    timestamps=np.arange(10).astype(float),
    electrodes=None,
):
    return SpikeEventSeries(
        name=name or name_generator("SpikeEventSeries"),
        description=description,
        data=data if data is not None else np.ones((10, 5)),
        timestamps=timestamps if timestamps is not None else np.arange(10).astype(float),
        electrodes=electrodes if electrodes is not None else mock_electrodes(),
    )



import numpy as np

from ...ogen import OptogeneticStimulusSite, OptogeneticSeries

from .device import mock_Device
from .utils import name_generator


def mock_OptogeneticStimulusSite(
    name=None,
    device=None,
    description="optogenetic stimulus site",
    excitation_lambda=500.,
    location="part of the brain",
):
    return OptogeneticStimulusSite(
        name=name or name_generator("OptogeneticStimulusSite"),
        device=device or mock_Device(),
        description=description,
        excitation_lambda=excitation_lambda,
        location=location
    )


def mock_OptogeneticSeries(
    name=None,
    data=None,
    site=None,
    resolution=-1.0,
    conversion=1.0,
    timestamps=None,
    starting_time=None,
    rate=10.0,
    comments="no comments",
    description="no description",
    control=None,
    control_description=None,
):
    return OptogeneticSeries(
        name=name or name_generator("OptogeneticSeries"),
        data=data if data is not None else np.array([1, 2, 3, 4]),
        site=site or mock_OptogeneticStimulusSite(),
        resolution=resolution,
        conversion=conversion,
        timestamps=timestamps,
        starting_time=starting_time,
        rate=rate,
        comments=comments,
        description=description,
        control=control,
        control_description=control_description,
    )



from unittest import TestCase

import pynwb


class TestImportStructure(TestCase):
    """Test whether the classes/modules imported from pynwb in version 2.1.1 are still accessible.

    NOTE: this test was needed to ensure backward compatibility of "import pynwb" after changes to the package file
    hierarchy in PyNWB 2.2.0 around validate.py (see https://github.com/NeurodataWithoutBorders/pynwb/pull/1511).
    """
    def test_outer_import_structure(self):
        current_structure = dir(pynwb)
        expected_structure = [
            "BuildManager",
            "CORE_NAMESPACE",
            "DataChunkIterator",
            "H5DataIO",
            "HDMFIO",
            "NWBContainer",
            "NWBData",
            "NWBDatasetSpec",
            "NWBFile",
            "NWBGroupSpec",
            "NWBHDF5IO",
            "NWBNamespace",
            "NamespaceCatalog",
            "Path",
            "ProcessingModule",
            "TimeSeries",
            "TypeMap",
            "_HDF5IO",
            "__NS_CATALOG",
            "__TYPE_MAP",
            "__builtins__",
            "__cached__",
            "__doc__",
            "__file__",
            "__get_resources",
            "__io",
            "__loader__",
            "__name__",
            "__package__",
            "__path__",
            "__resources",
            "__spec__",
            "__version__",
            "_due",
            "_get_resources",
            "_version",
            "available_namespaces",
            "base",
            "behavior",
            "core",
            "deepcopy",
            "device",
            "docval",
            "ecephys",
            "epoch",
            "file",
            "get_class",
            "get_docval",
            "get_manager",
            "get_type_map",
            "getargs",
            "h5py",
            "hdmf",
            "hdmf_typemap",
            "icephys",
            "image",
            "io",
            "legacy",
            "load_namespaces",
            "misc",
            "ogen",
            "ophys",
            "os",
            "popargs",
            "register_class",
            "register_map",
            "retinotopy",
            "spec",
            "testing",
            "validate",
            "warn",
        ]
        for member in expected_structure:
            self.assertIn(member=member, container=current_structure)



"""
The mock module provides mock instances of common neurodata types which can be used to write
tests for downstream libraries. For instance, to write an RoiResponseSeries, you need a
PlaneSegmentation, which requires an ImagingPlane, which in turn requires an
OpticalChannel and a Device, all of which need to be populated with reasonable mock
data. This library streamlines the process of creating mock objects by generating the
required prerequisites for this object on-the-fly and automatically using reasonable
defaults. Any of these default objects and parameters can be overridden.
"""



import warnings

import numpy as np

from pynwb.ecephys import ElectricalSeries, SpikeEventSeries, EventDetection, Clustering, EventWaveform,\
                          ClusterWaveforms, LFP, FilteredEphys, FeatureExtraction, ElectrodeGroup
from pynwb.device import Device
from pynwb.file import ElectrodeTable
from pynwb.testing import TestCase

from hdmf.common import DynamicTableRegion


def make_electrode_table():
    table = ElectrodeTable()
    dev1 = Device('dev1')
    group = ElectrodeGroup('tetrode1', 'tetrode description', 'tetrode location', dev1)
    table.add_row(location='CA1', group=group, group_name='tetrode1')
    table.add_row(location='CA1', group=group, group_name='tetrode1')
    table.add_row(location='CA1', group=group, group_name='tetrode1')
    table.add_row(location='CA1', group=group, group_name='tetrode1')

    return table


class ElectricalSeriesConstructor(TestCase):

    def _create_table_and_region(self):
        table = make_electrode_table()
        region = DynamicTableRegion(
            name='electrodes',
            data=[0, 2],
            description='the first and third electrodes',
            table=table
        )
        return table, region

    def test_init(self):
        data = list(range(10))
        ts = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
        channel_conversion = [2., 6.3]
        filtering = 'Low-pass filter at 300 Hz'
        table, region = self._create_table_and_region()
        eS = ElectricalSeries(
            name='test_eS',
            data=data,
            electrodes=region,
            channel_conversion=channel_conversion,
            filtering=filtering,
            timestamps=ts
        )
        self.assertEqual(eS.name, 'test_eS')
        self.assertEqual(eS.data, data)
        self.assertEqual(eS.electrodes, region)
        self.assertEqual(eS.timestamps, ts)
        self.assertEqual(eS.channel_conversion, [2., 6.3])
        self.assertEqual(eS.filtering, filtering)

    def test_link(self):
        table, region = self._create_table_and_region()
        ts1 = ElectricalSeries('test_ts1', [0, 1, 2, 3, 4, 5], region, timestamps=[0.0, 0.1, 0.2, 0.3, 0.4, 0.5])
        ts2 = ElectricalSeries('test_ts2', ts1, region, timestamps=ts1)
        ts3 = ElectricalSeries('test_ts3', ts2, region, timestamps=ts2)
        self.assertEqual(ts2.data, [0, 1, 2, 3, 4, 5])
        self.assertEqual(ts2.timestamps, [0.0, 0.1, 0.2, 0.3, 0.4, 0.5])
        self.assertEqual(ts3.data, [0, 1, 2, 3, 4, 5])
        self.assertEqual(ts3.timestamps, [0.0, 0.1, 0.2, 0.3, 0.4, 0.5])

    def test_invalid_data_shape(self):
        table, region = self._create_table_and_region()
        with self.assertRaisesWith(ValueError, ("ElectricalSeries.__init__: incorrect shape for 'data' (got '(2, 2, 2, "
                                                "2)', expected '((None,), (None, None), (None, None, None))')")):
            ElectricalSeries('test_ts1', np.ones((2, 2, 2, 2)), region, timestamps=[0.0, 0.1, 0.2, 0.3, 0.4, 0.5])

    def test_dimensions_warning(self):
        table, region = self._create_table_and_region()
        with warnings.catch_warnings(record=True) as w:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")
            ElectricalSeries(
                name="test_ts1",
                data=np.ones((6, 3)),
                electrodes=region,
                timestamps=[0.0, 0.1, 0.2, 0.3, 0.4, 0.5]
            )
            self.assertEqual(len(w), 1)
            assert (
                "ElectricalSeries 'test_ts1': The second dimension of data does not match the length of electrodes. "
                "Your data may be transposed."
                ) in str(w[-1].message)

        with warnings.catch_warnings(record=True) as w:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")
            ElectricalSeries(
                name="test_ts1",
                data=np.ones((2, 6)),
                electrodes=region,
                rate=30000.,
            )
            self.assertEqual(len(w), 1)
            assert (
               "ElectricalSeries 'test_ts1': The second dimension of data does not match the length of electrodes, "
               "but instead the first does. Data is oriented incorrectly and should be transposed."
                   ) in str(w[-1].message)


class SpikeEventSeriesConstructor(TestCase):

    def _create_table_and_region(self):
        table = make_electrode_table()
        region = DynamicTableRegion(
            name='electrodes',
            data=[1, 3],
            description='the second and fourth electrodes',
            table=table
        )
        return table, region

    def test_init(self):
        table, region = self._create_table_and_region()
        data = ((1, 1), (2, 2), (3, 3))
        timestamps = np.arange(3)
        sES = SpikeEventSeries(name='test_sES', data=data, timestamps=timestamps, electrodes=region)
        self.assertEqual(sES.name, 'test_sES')
        # self.assertListEqual(sES.data, data)
        np.testing.assert_array_equal(sES.data, data)
        np.testing.assert_array_equal(sES.timestamps, timestamps)

    def test_no_rate(self):
        table, region = self._create_table_and_region()
        data = ((1, 1, 1), (2, 2, 2))
        with self.assertRaises(TypeError):
            SpikeEventSeries(name='test_sES', data=data, electrodes=region, rate=1.)

    def test_incorrect_timestamps(self):
        table, region = self._create_table_and_region()
        data = ((1, 1, 1), (2, 2, 2))
        with self.assertRaisesWith(ValueError, "Must provide the same number of timestamps and spike events"):
            SpikeEventSeries(name='test_sES', data=data, electrodes=region, timestamps=[1.0, 2.0, 3.0])


class ElectrodeGroupConstructor(TestCase):

    def test_init(self):
        dev1 = Device('dev1')
        group = ElectrodeGroup('elec1', 'electrode description', 'electrode location', dev1, (1, 2, 3))
        self.assertEqual(group.name, 'elec1')
        self.assertEqual(group.description, 'electrode description')
        self.assertEqual(group.location, 'electrode location')
        self.assertEqual(group.device, dev1)
        self.assertEqual(group.position, (1, 2, 3))

    def test_init_position_none(self):
        dev1 = Device('dev1')
        group = ElectrodeGroup('elec1', 'electrode description', 'electrode location', dev1)
        self.assertEqual(group.name, 'elec1')
        self.assertEqual(group.description, 'electrode description')
        self.assertEqual(group.location, 'electrode location')
        self.assertEqual(group.device, dev1)
        self.assertIsNone(group.position)

    def test_init_position_bad(self):
        dev1 = Device('dev1')
        with self.assertRaises(ValueError):
            ElectrodeGroup('elec1', 'electrode description', 'electrode location', dev1, (1, 2))


class EventDetectionConstructor(TestCase):

    def _create_table_and_region(self):
        table = make_electrode_table()
        region = DynamicTableRegion(
            name='electrodes',
            data=[0, 2],
            description='the first and third electrodes',
            table=table
        )
        return table, region

    def test_init(self):
        data = list(range(10))
        ts = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
        table, region = self._create_table_and_region()
        eS = ElectricalSeries('test_eS', data, region, timestamps=ts)
        eD = EventDetection('detection_method', eS, (1, 2, 3), (0.1, 0.2, 0.3))
        self.assertEqual(eD.detection_method, 'detection_method')
        self.assertEqual(eD.source_electricalseries, eS)
        self.assertEqual(eD.source_idx, (1, 2, 3))
        self.assertEqual(eD.times, (0.1, 0.2, 0.3))
        self.assertEqual(eD.unit, 'seconds')


class EventWaveformConstructor(TestCase):

    def _create_table_and_region(self):
        table = make_electrode_table()
        region = DynamicTableRegion(
            name='electrodes',
            data=[0, 2],
            description='the first and third electrodes',
            table=table
        )
        return table, region

    def test_init(self):
        table, region = self._create_table_and_region()
        sES = SpikeEventSeries('test_sES', list(range(10)), list(range(10)), region)

        ew = EventWaveform(sES)
        self.assertEqual(ew.spike_event_series['test_sES'], sES)
        self.assertEqual(ew['test_sES'], ew.spike_event_series['test_sES'])


class ClusteringConstructor(TestCase):

    def test_init(self):
        times = [1.3, 2.3]
        num = [3, 4]
        peak_over_rms = [5.3, 6.3]

        with self.assertWarnsWith(DeprecationWarning, 'use pynwb.misc.Units or NWBFile.units instead'):
            cc = Clustering(description='description', num=num, peak_over_rms=peak_over_rms, times=times)
        self.assertEqual(cc.description, 'description')
        self.assertEqual(cc.num, num)
        self.assertEqual(cc.peak_over_rms, peak_over_rms)
        self.assertEqual(cc.times, times)


class ClusterWaveformsConstructor(TestCase):

    def test_init(self):
        times = [1.3, 2.3]
        num = [3, 4]
        peak_over_rms = [5.3, 6.3]
        with self.assertWarnsWith(DeprecationWarning, 'use pynwb.misc.Units or NWBFile.units instead'):
            cc = Clustering(description='description', num=num, peak_over_rms=peak_over_rms, times=times)

        means = [[7.3, 7.3]]
        stdevs = [[8.3, 8.3]]

        with self.assertWarnsWith(DeprecationWarning, 'use pynwb.misc.Units or NWBFile.units instead'):
            cw = ClusterWaveforms(
                clustering_interface=cc,
                waveform_filtering='filtering',
                waveform_mean=means,
                waveform_sd=stdevs
            )
        self.assertEqual(cw.clustering_interface, cc)
        self.assertEqual(cw.waveform_filtering, 'filtering')
        self.assertEqual(cw.waveform_mean, means)
        self.assertEqual(cw.waveform_sd, stdevs)


class LFPTest(TestCase):

    def _create_table_and_region(self):
        table = make_electrode_table()
        region = DynamicTableRegion(
            name='electrodes',
            data=[0, 2],
            description='the first and third electrodes',
            table=table
        )
        return table, region

    def test_add_electrical_series(self):
        lfp = LFP()
        table, region = self._create_table_and_region()
        eS = ElectricalSeries('test_eS', [0, 1, 2, 3], region, timestamps=[0.1, 0.2, 0.3, 0.4])
        lfp.add_electrical_series(eS)
        self.assertEqual(lfp.electrical_series.get('test_eS'), eS)


class FilteredEphysTest(TestCase):

    def _create_table_and_region(self):
        table = make_electrode_table()
        region = DynamicTableRegion(
            name='electrodes',
            data=[0, 2],
            description='the first and third electrodes',
            table=table
        )
        return table, region

    def test_init(self):
        table, region = self._create_table_and_region()
        eS = ElectricalSeries('test_eS', [0, 1, 2, 3], region, timestamps=[0.1, 0.2, 0.3, 0.4])
        fe = FilteredEphys(eS)
        self.assertEqual(fe.electrical_series.get('test_eS'), eS)
        self.assertEqual(fe['test_eS'], fe.electrical_series.get('test_eS'))

    def test_add_electrical_series(self):
        fe = FilteredEphys()
        table, region = self._create_table_and_region()
        eS = ElectricalSeries('test_eS', [0, 1, 2, 3], region, timestamps=[0.1, 0.2, 0.3, 0.4])
        fe.add_electrical_series(eS)
        self.assertEqual(fe.electrical_series.get('test_eS'), eS)
        self.assertEqual(fe['test_eS'], fe.electrical_series.get('test_eS'))


class FeatureExtractionConstructor(TestCase):

    def _create_table_and_region(self):
        table = make_electrode_table()
        region = DynamicTableRegion(
            name='electrodes',
            data=[0, 2],
            description='the first and third electrodes',
            table=table
        )
        return table, region

    def test_init(self):
        event_times = [1.9, 3.5]
        table, region = self._create_table_and_region()
        description = ['desc1', 'desc2', 'desc3']
        features = [[[0, 1, 2], [3, 4, 5]], [[6, 7, 8], [9, 10, 11]]]
        fe = FeatureExtraction(region, description, event_times, features)
        self.assertEqual(fe.description, description)
        self.assertEqual(fe.times, event_times)
        self.assertEqual(fe.features, features)

    def test_invalid_init_mismatched_event_times(self):
        event_times = []  # Need 1 event time but give 0
        table, region = self._create_table_and_region()
        description = ['desc1', 'desc2', 'desc3']
        features = [[[0, 1, 2], [3, 4, 5]]]
        self.assertRaises(ValueError, FeatureExtraction, region, description, event_times, features)

    def test_invalid_init_mismatched_electrodes(self):
        event_times = [1]
        table = make_electrode_table()
        region = DynamicTableRegion(name='electrodes', data=[0], description='the first electrode', table=table)
        description = ['desc1', 'desc2', 'desc3']
        features = [[[0, 1, 2], [3, 4, 5]]]
        self.assertRaises(ValueError, FeatureExtraction, region, description, event_times, features)

    def test_invalid_init_mismatched_description(self):
        event_times = [1]
        table, region = self._create_table_and_region()
        description = ['desc1', 'desc2', 'desc3', 'desc4']  # Need 3 descriptions but give 4
        features = [[[0, 1, 2], [3, 4, 5]]]
        self.assertRaises(ValueError, FeatureExtraction, region, description, event_times, features)

    def test_invalid_init_mismatched_description2(self):
        event_times = [1]
        table, region = self._create_table_and_region()
        description = ['desc1', 'desc2', 'desc3']
        features = [[0, 1, 2], [3, 4, 5]]  # Need 3D feature array but give only 2D array
        self.assertRaises(ValueError, FeatureExtraction, region, description, event_times, features)



import numpy as np

from ...base import TimeSeries
from .utils import name_generator


def mock_TimeSeries(
    name=None,
    data=None,
    unit="volts",
    resolution=-1.0,
    conversion=1.0,
    timestamps=None,
    starting_time=None,
    rate=10.0,
    comments="no comments",
    description="no description",
    control=None,
    control_description=None,
    continuity=None,
):
    return TimeSeries(
        name=name or name_generator("TimeSeries"),
        data=data if data is not None else np.array([1, 2, 3, 4]),
        unit=unit,
        resolution=resolution,
        conversion=conversion,
        timestamps=timestamps,
        starting_time=starting_time,
        rate=rate,
        comments=comments,
        description=description,
        control=control,
        control_description=control_description,
        continuity=continuity,
    )
